/*===========================================================================*/
/* Library of support routines                                               */
/*===========================================================================*/
#define CCDUTILS_C

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <memory.h>
#include <malloc.h>
#include <sys/types.h>

#include "cdl.h"
#include "motxer.h"

/*========== Function prototypes ============================================*/


/*========== Parse routines ================================================*/
#define MAXTOKEN 40

static char *tokenbuf[MAXTOKEN];    /* array of pointers to tokens */
static char **tokenv;               /* pointer to current token */


int
fp_tokenize(char *str)
{
    extern char *tokenbuf[];
    extern char **tokenv;
    int i=0;                        /* current token (index) */

    tokenbuf[i++] = strtok(str," \n\r");
    while ((tokenbuf[i] = strtok(NULL," \n\r")) != NULL) i++;

    tokenbuf[i] = 00 ;
    tokenv = &tokenbuf[0];
}

/* parse a command line */
int 
fp_parse(CMD_ENTRY cmd_dict[], int dict_size)
{
    int entry=0, n;
    char *token;

    while((token = nxttoken()) != 0 ) {
        if (strncmp(token,"#",1)==0) break;

        /* look for the current token in the command table */
        for (entry=0; entry < dict_size; entry++ ) {
            n = strlen(cmd_dict[entry].nmem);
            if(strncmp(token, cmd_dict[entry].nmem, n)==0) break;
        }

        /* if found, execute cmd function */
        if (entry < dict_size) 
            (cmd_dict[entry].fcn)();    
        else {
            printf("* * * *   %s  not found   * * * *\n",token);
            break;
        }
    }
}

/* increment token vector & supply pointer or 00 */
char *
nxttoken() 
{
    extern char **tokenv;

    if(*tokenv != 0) 
        return(*tokenv++);
    else
        return(0);
}

/* return a floating point arg from the token list */
double
fp_arg()
{
    char *nxtarg = nxttoken();

    if (nxtarg == NULL) 
        return(0.0);
    else
        return(atof(nxtarg));

}

int
scriptrun(char* fname)
{
    extern CMD_ENTRY cmd_dict[];
    extern int dict_size;
    char  fpath[80];
    char  cmdbuf[80];
    FILE  *fp;
    int r;

    if (fname == NULL) return(ABORT);
    r = find_file(fname, fpath);
    if (r < 0) return(ABORT);
    fp = fopen(fpath, "r");

    while (fgets(cmdbuf, 80, fp) != NULL) { 
        printf("%s", cmdbuf);
        fp_tokenize(cmdbuf);
        fp_parse(cmd_dict, dict_size);
    }

    fclose(fp);
    return(OK);
}

/* look for a file in the usual places */
int
find_file(char* fname, char* fpath)
{
    int fd;

    if (fname != NULL) {                /* look in current directory */
        strcpy(fpath, fname);
        fd = open(fpath, O_RDONLY);
    }

    if(fd < 0) {                        /* look in home directory */
        sprintf(fpath, "%s/%s", getenv("HOME"), fname);
        fd = open(fpath, O_RDONLY);
    }

    if (fd < 0) {
        printf("File not found: %s\n", fname);
        return(ABORT);
    }
    close(fd);
    return(OK);
}


/*===========================================================================*/
/* Display device routines  (for iis model 70/f                              */
/*===========================================================================*/
static CDLPtr cdl;

static int dd_zmin      = -1;
static int dd_zmax      = -1;
static int dd_zspan     = -1;
static int dd_config    = 3;

int
ddopen()
{
    extern CDLPtr cdl;

    cdl = cdl_open ("unix:/tmp/.IMT%d" );
    if (cdl==NULL) {
        return(0);          // was exit(-1);
    }
    return(1);

//    /* Open the package and a connection to the server. */
//    if (!(cdl = cdl_open ("unix:/tmp/.IMT%d" ))) exit (-1);
//    if (!(cdl = cdl_open ("inet:5137" ))) exit (-1);

}

void
ddclose()
{
    extern CDLPtr cdl;

    cdl_close (cdl);
}

void
dderase()
{
    extern CDLPtr cdl;

    cdl_clearFrame (cdl);
}

int
ddwriteimage(IMDESC *im)
{
    extern CDLPtr cdl;
    extern int *dispbuf;
    int status=0, frame=1;
    int nx = im->ncol;
    int ny = im->nrow;
    int j;

    strcpy(im->objtitle,"object");

    for (j=0; j<nx*ny; j++){
        dispbuf[j] = (int)(im->pixbuf[j]);
    }
    if (cdl_displayPix(cdl,(uchar *)dispbuf, nx, ny, 32, frame, FB_AUTO, 1)){
        status = 1;
    }

    return (status);
}

int
ddsetconfig(int config_num)
{
    extern int dd_config;

    dd_config = config_num;
}

int
ddsetzrange(int min, int max)
{
    extern int dd_zmin, dd_zmax;

    dd_zmin = min;
    dd_zmax = max;
}

int
ddsetzmin(int min)
{
    extern int dd_zmin;

    dd_zmin = min;
}

int
ddsetzspan(int span)
{
    extern int dd_zspan;

    dd_zspan = span;
}

int
ddsetzmax(int max)
{
    extern int dd_zmax;

    dd_zmax = max;
}


/*===========================================================================*/
/* Fitslib routines                                                          */
/*===========================================================================*/
#define  KEY_LEN            8           /* Length of FITS keyword */

#define  FITS_REC_SIZE      (36*80)     /* FITS record size */
#define  FITS_CARD_SIZE     80          /* FITS header card size */
#define  FITS_NCARDS        36          /* Number of cards in record */

#define  MAXAXIS            3           /* Max number of axes allowed */
#define  MAXDIM             5000        /* Max dimension of each axis */
#define  MAXSHORT           32768       /* Max value of short int */

/* Define formats used to setup the FITS header records */
#define  BL_FMT         "%-8.8s= %20s  /  %-45.45s"
#define  INT_FMT        "%-8.8s= %20d  /  %-45.45s"
#define  DBL_FMT        "%-8.8s= %20.3f  /  %-45.45s"
#define  STR_FMT        "%-8.8s= \'%-18.18s\'  /  %-45.45s"
#define  COM_FMT        "%-8.8s  %-70.70s"
#define  END_FMT        "%-80.80s"

typedef char FITSCARD[FITS_CARD_SIZE+1];
// static int max_data = 32768;                /* data is 14-bit */

/* Display an image descriptor contents */
void  
imdesc_prt(IMDESC *im)
{
    printf("SEQMODE   = %s\n", im->seq_mode);
    printf("NCOL      = %d\n", im->ncol);
    printf("NROW      = %d\n", im->nrow);
    printf("NCH       = %d\n", im->nch);
    printf("NCCD      = %d\n", im->nccd);

    printf("OBJECT    = %s\n", im->objtitle); 
    printf("DATEOBS   = %s\n", im->dateobs);
    printf("UTIME     = %s\n", im->utime);
    printf("EXPTIME   = %s\n", im->exptime);
}

/* Initialize an image descriptor structure  */
void  
imdesc_init(IMDESC *im)
{
/*  im->pixbuf = 0;     should check for a buffer to free */
    im->bitpix = 16;
    im->naxis  = 2;
    im->ncol   = 0;
    im->nrow   = 0;
    im->nch    = 1; 
                                                         
    strcpy(im->seq_mode,    "no_instrument");
    strcpy(im->dateobs,     "2000-01-01");
    strcpy(im->utime,       "00:00:00.00");
    strcpy(im->exptime,     "00:00:00.00");
    strcpy(im->objtitle,    "no_object");
}

/* Read a FITS image file from disk  */
int 
ccdrdfits(IMDESC *im, char *fname)
{
    int  i, r, fp;
    int  npix;
    int  nbytes;                        /* number of bytes in image */
    int  end_stat = 0;                  /* END card flag */
    int  ncard = 0;                     /* number of cards to read */
    double  bzero, bscale;
    char card[FITS_CARD_SIZE+1];        /* storage for FITS header cards */
    char *valstr, *locate, fits_true;   /* pointers to header fields */
    char *pbuf=(char *)im->pixbuf;

    if ((fp = open(fname, O_RDONLY)) == -1){
        printf("could not open file: %s\n", fname); 
        return(ABORT); 
    }

    imdesc_init(im); 

    /* Read and decode the fits header record into an image descriptor */
    /* Only one header block is expected -- ie max 36 FITS cards */

     while (ncard<FITS_NCARDS || end_stat == 0) {
        ncard++ ;               

        r = read(fp, card, FITS_CARD_SIZE);
        if(r != FITS_CARD_SIZE) {
            printf("Error reading header card %d \n", ncard);
        }
        card[FITS_CARD_SIZE] = (char)0; 

        /* print FITS card to display monitor */
        if (end_stat == 0) printf("%-65.65s\n", card); 

        /* Decode the variable field, if present */
        locate = card + KEY_LEN;
        if(*locate++ == '=') {      
            while(*locate == ' ') locate++ ;    /* move to variable field */

            if(*locate == '\00') {          /* check for end of card */
                printf("End of card before variable found \n");
                valstr = locate;
            }
            else if(*locate == '\'')    {       /* must be string variable */
                valstr = ++locate;
                locate = strchr(valstr,'\'');
            }
            else {                      /* must be numeric variable */
                valstr = locate;
                locate = strchr(valstr,' ');
            }
            *locate = (char)0;
        }

        /* Decode the keyword field */ 
            if(strncmp(card,"SIMPLE  ",KEY_LEN) == 0) {
            fits_true = *valstr;
        } 
        else if(strncmp(card,"NAXIS   ",KEY_LEN) == 0) {
            im->naxis = atoi(valstr);
            if(im->naxis == 2)  im->nch = 1; 
        } 
        else if(strncmp(card,"NAXIS1  ",KEY_LEN) == 0) {
            im->ncol = atoi(valstr);
        } 
        else if(strncmp(card,"NAXIS2  ",KEY_LEN) == 0) {
            im->nrow = atoi(valstr);
        } 
        else if(strncmp(card,"NAXIS3  ",KEY_LEN) == 0){
            im->nch = atoi(valstr);
        } 
        else if(strncmp(card,"BSCALE  ",KEY_LEN) == 0) {
            bscale = atof(valstr);
        }
        else if(strncmp(card,"BZERO   ",KEY_LEN) == 0) {
            bzero = atof(valstr);
        }
        else if (strncmp(card,"SEQ_MODE",KEY_LEN) == 0) {
            strcpy(im->seq_mode, valstr);
        }
        else if (strncmp(card,"OBJECT  ",KEY_LEN) == 0) {
            strcpy(im->objtitle, valstr);
        } 
        else if (strncmp(card,"DATE-OBS",KEY_LEN) == 0) {
            strcpy(im->dateobs, valstr);
        } 
        else if (strncmp(card,"UTIME   ",KEY_LEN) == 0) {
            strcpy(im->utime, valstr);
        } 
        else if (strncmp(card,"EXPTIME ",KEY_LEN) == 0) {
            strcpy(im->exptime, valstr);
        } 
        else if (strncmp(card,"END     ",KEY_LEN) == 0) {
            end_stat = 1;
        } 
    }

    /* Elementary error checking on the decoded header variables */
    if(fits_true != 'T') printf("Non-standard FITS format\n");
    if(im->naxis > MAXAXIS)printf("FITS NAXIS too large");
    if(im->ncol  > MAXDIM) printf("FITS NAXIS1 too large\n"); 
    if(im->nrow  > MAXDIM) printf("FITS NAXIS2 too large\n");
    if(im->nch   > MAXDIM) printf("FITS NAXIS3 too large\n");

    /* read pixel information from the FITS image file */
    npix = (im->ncol)*(im->nrow);
    nbytes = 2*npix;
    if(im->pixbuf == NULL) {
        printf("pix buffer not initialized -- null pointer\n");
        return(ABORT);
    }

    r = read(fp, im->pixbuf, nbytes);   
    if(r != nbytes) {
        printf("Error reading pixel data, r = %d\n", r);
        close(fp);
        return(ABORT);
    }

     /* Add bzero if necessary and swap bytes */
    if (!BIGxENDIAN)swab(pbuf,pbuf,npix*sizeof(u_short));
    for (i=0; i<npix; i++){
        im->pixbuf[i] += (int)bzero;
    }

    close(fp);
    return(OK);
}

/* Make a FITS image file  */
int  
ccdwrfits(IMDESC *im, char *fname)
{
    FITSCARD fits_header[FITS_NCARDS], *buf=fits_header;
    char fbuf[FITS_REC_SIZE] ;
    int i, j, n=0, pc, r, npix, nbytes, nfill ;
    short pbuf[1024];
    int fp;

    if ((fp = open(fname, O_RDWR|O_CREAT, 0666)) == -1){
        printf("could not open file: %s\n", fname); 
        return(ABORT); 
    }

    /* Initialize the fits header cardimages to all spaces */
    for(i=0; i<FITS_NCARDS; i++) {
        memset(fits_header[i],(char)040,FITS_CARD_SIZE);
        fits_header[i][FITS_CARD_SIZE] = (char)0;
    }

    /* Store keywords in an IRAF image header & write to disk */
    sprintf((void *)buf++, BL_FMT,  "SIMPLE", "T", "FITS STANDARD");
    sprintf((void *)buf++, INT_FMT, "BITPIX", im->bitpix, "BITS/PIXEL");
    sprintf((void *)buf++, INT_FMT, "NAXIS" , im->naxis, "NUMBER OF AXES");
    sprintf((void *)buf++, INT_FMT, "NAXIS1", im->ncol, "");
    sprintf((void *)buf++, INT_FMT, "NAXIS2", im->nrow, "");
    sprintf((void *)buf++, INT_FMT, "NAXIS3", im->nch, ""); 

    sprintf((void *)buf++, DBL_FMT, "BSCALE", 1.0, "");
    sprintf((void *)buf++, DBL_FMT, "BZERO ", (float)MAXSHORT, "");

    sprintf((void *)buf++, STR_FMT, "SEQ_MODE", im->seq_mode, "");
    sprintf((void *)buf++, STR_FMT, "OBJECT  ", im->objtitle, "object name");
    sprintf((void *)buf++, STR_FMT, "DATE-OBS", im->dateobs, "(yyyy-mm-dd)");
    sprintf((void *)buf++, STR_FMT, "UTIME   ", im->utime, "UT @ start");
    sprintf((void *)buf++, STR_FMT, "EXPTIME ", im->exptime, "exposure time");

    sprintf((void *)buf++, END_FMT, "END     ");

    /* Write fits header to disk */
    for(i=0; i<FITS_NCARDS; i++) {  
        r = write(fp, fits_header[i], FITS_CARD_SIZE);
        if(r != FITS_CARD_SIZE) {
            printf("Error writing headercard %d\n", i);
            return(ABORT);
        }
    }
    
    /* Write the image data and zero fill the remainder */
    npix = (im->ncol) * (im->nrow);
    nbytes  = 2 * npix;
    nfill = FITS_REC_SIZE - (nbytes % FITS_REC_SIZE);

    while (n<npix){
        for (pc=0; pc<1024; pc++,n++){
            pbuf[pc] = (short)((int)(im->pixbuf[n])-MAXSHORT);
            if (n >= npix) break;
        }
        if (!BIGxENDIAN)swab(pbuf,pbuf,pc*sizeof(u_short));
        write(fp,pbuf,2*pc);
    }
    r = write(fp, fbuf, nfill);

    close(fp);
    return(OK); 
}

/*==================== End Of File ========================================= */
