/*===========================================================================*/
/* Routines in the command dictionary                                        */
/*===========================================================================*/
#define CMDDICT_C

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <malloc.h>
#include <sys/types.h>
#include <time.h>
#include "motxer.h"

/*========== command dictionary =============================================*/
CMD_ENTRY cmd_dict[] =
    {
        { "@@",         cx_scriptrun },
        { "abort",      ccdexp_abort },
        { "amp",        hh_set_opch },
        { "autoflush",  set_autoflush },
        { "autopid",    hh_set_pid },
        { "autosave",   set_autosave },
        { "autotemp",   set_autotemp },
        { "autowipe",   set_autowipe },
        { "cds",        hh_set_usec },
        { "ch",         hh_set_opch },
        { "cmd",        get_cmdline },
        { "config",     set_ddconfig },
        { "cycle",      set_cycle },
        { "cyctime",    set_cyctime },
        { "echo",       echo },
/*      { "exit",       cx_exit }, */
        { "exp",        set_exptimer },
        { "extend",     tx_extend },
        { "flush",      set_autoflush },
        { "fprefix",    set_fprefix },
        { "fsave",      set_nfsave },
        { "gain",       hh_set_gain },
        { "heat",       hh_set_heat },
        { "hbin",       hh_set_hbin },
        { "impath",     set_impath },
        { "init",       seq_init },
        { "load",       seq_initII },
        { "namps",      hh_set_namps },
        { "nccd",       hh_set_nccd },
        { "ncols",      hh_set_ncols },
        { "nrows",      hh_set_nrows },
        { "opch",       hh_set_opch },
        { "pca",        plot_coladd },
        { "pco",        plot_col },
        { "pid",        hh_set_pid },
        { "pra",        plot_rowadd },
        { "pro",        plot_row },
        { "rdline",     set_rdflag },
        { "read",       ccdexp_terminate },
        { "save",       save_fitsimage },
        { "sclks",      hh_set_sclks },
        { "show",       show_image },
        { "shutter",    set_autoshutter },
        { "start",      ccdexp_start },
        { "stat",       hh_exp_status },
        { "tec",        hh_set_heat },
        { "temp",       set_autotemp },
        { "usec",       hh_set_usec },
        { "vbin",       hh_set_vbin },
        { "vget",       hh_get_vdacs },
        { "wipe",       set_autowipe },
        { "xrange",     set_xrange },
        { "yrange",     set_yrange },
        { "zmin",       set_zmin },
        { "zmax",       set_zmax },
        { "zrange",     set_zrange },
        { "zspan",      set_zspan }
    };

int dict_size = (sizeof(cmd_dict) / sizeof(CMD_ENTRY));



/*========== Subroutines ====================================================*/
void
seq_init(void)
{
    extern SEQ_MODE *seq;
    char *fname, fpath[80];
    int r;

    /* Get the filename and search for sequencer code file */
    fname = nxttoken();
    if( strchr(fname,'.')==NULL ) strcat(fname,".s");
    if(fname == NULL || find_file(fname, fpath) < 0) {
        printf("Sequencer codefile not found\n");
        return;
    }

    printf("Opening sequencer codefile %s\n",fpath);
    r = dsp_init(fpath);
    if (r>0)seq_init_update();
}

void
seq_initII(void)
{
    extern SEQ_MODE *seq;
    char *fname, fpath[80];
    int r;

    /* Get the filename and search for sequencer code file */
    fname = nxttoken();
    if( strchr(fname,'.')==NULL ) strcat(fname,".s");
    if(fname == NULL || find_file(fname, fpath) < 0) {
        printf("Sequencer codefile not found\n");
        return;
    }

    printf("Opening sequencer codefile %s\n",fpath);
    r = dsp_initII(fpath);
    if (r>0)seq_init_update();
}

void
seq_init_update(void)
{
    extern SEQ_MODE *seq;
    extern int pidflag;
    extern int autotemp;
    extern double exptime;
    int r;

    printf("Getting parameters from DSP\n");
    seq->nrow   = get_nrow();
    seq->ncol   = get_ncol();
    seq->nft    = get_nft();
    seq->nflush = get_nflush();
    seq->nccd   = get_nccd();
//    seq->namps  = get_namps();
    seq->vbin   = get_vbin();
    seq->hbin   = get_hbin();
    seq->vskip  = get_vskip();
    seq->hskip  = get_hskip();
//    seq->vsub   = get_vsub();
//    seq->hsub   = get_hsub();
    seq->nshuffle = get_nshuffle();
    seq->seqtime  = get_seqtime();
    seq->gain   = get_gain();
    seq->usec   = get_usec();
    seq->opch   = get_opch();
    seq->sclks  = get_sclks();
    seq->pid    = get_pid();

// Need to put in some elementary error checking
#ifdef GCAM
    printf("Getting DAC voltages from DSP\n");
    gcam_get_vdac(seq->vbias,seq->vclk);
    gcam_get_temp(seq->temp);
#endif // GCAM

#ifdef BASE
    printf("Getting DAC voltages from DSP\n");
    base_get_vdac(seq->vbias,seq->vclk,seq->opch);
    base_get_heat(seq->heat);
    base_get_tset(seq->tset);
    base_get_temp(seq->temp);
#endif // BASE

    printf("Updating display\n");
    ud_nrows(seq->nrow);
    ud_ncols(seq->ncol);
    ud_nccd(seq->nccd);
    ud_vbin(seq->vbin);
    ud_hbin(seq->hbin);
    ud_gain(seq->gain);
    ud_opch(seq->opch);
    ud_sclks(seq->sclks);
    ud_usec(seq->usec);
    ud_temp();
    ud_heat();
    ud_hhpanel();
//    ud_infopanel();
    
    ud_autotemp(autotemp);
    pidflag = seq->pid;
    ud_autopid(pidflag);

    exptime = (seq->seqtime)/1000.0;
    ud_xtime();
}

void
ccdexp_start(void)
{
    extern SEQ_MODE  *seq;
    extern int expflag, cycleflag;

    /* don't start unless seqencer code has been loaded */
    if(seq == NULL) {
        printf("Sequencer modes not defined \n");
        return;
    }

    if(expflag == IDLE){
        printf("Exposure started\n");
        expflag = EXPSTART; 
    }
    else {
        printf("Exposure in progress\007\n");
        ud_cycle(CYCLE_OFF);
    }
    ud_exp_state();
}

void
ccdexp_terminate(void)
{
    extern int expflag;
    extern double prgtime, endtime, expopen, systime;

    if (expflag != EXPOSE) expopen=systime;

    endtime = prgtime + 0.01;
    expflag = EXPOSE;
    ud_exp_state();
}

void
ccdexp_abort(void)
{
    extern int expflag;

    if (expflag != IDLE) expflag = EXPABORT;
    printf("Exposure aborted \n");
    ud_exp_state();
}

void 
clear_ccd(void)
{
    int r;

    r = flush_ccd();
    if (r == 0) printf("Detector flushed \n");
    return;
}

/*  Set up and start the ccd exposure */
int
ccd_expose(void)
{
    extern SEQ_MODE *seq;
    extern IMDESC *im;
    extern char version[];
    extern u_short *imbuf;
    extern int *dispbuf;
    extern int imsize, xbsize, xbnum;
    extern int autoshutter, autowipe, autoflush;
    extern double expopen, expshut;
    int i, r=0;
    time_t  uxt = (time_t)expshut;

    /* Set the image descriptor information */
    im->naxis = 2;
    im->ncol = seq->ncol*seq->namps;
    im->nrow = seq->nrow;
    im->nccd = seq->nccd;
//    strcpy(im->seq_mode, "DSP 56303 v2.1");
    strcpy(im->seq_mode, version);
    strftime(im->utime, 20, "%T", gmtime(&uxt));
    sprintf(im->utime+8, ".%d", (int)(100.0*fmod(expshut,1.0)) );
    strftime(im->dateobs, 20, "%Y-%m-%d", gmtime(&uxt));
    sprintf(im->exptime, "%.2f", expshut - expopen);

    /* Set up the image buffers */
    if(imsize<(2*seq->nrow*seq->ncol*seq->namps) || xbnum<(seq->nccd*seq->namps)){
        free(imbuf);
        free(dispbuf);
        for (i=0; i<seq->nccd*seq->namps; i++)free(im->pix[i]);

        imsize = (2*seq->nrow*seq->ncol*seq->namps);
        imbuf = (u_short *)malloc(imsize);
        printf("Image buffer size %d kB\n",((imsize)>>10));

        dispbuf = (int *)malloc(2*imsize);
        xbsize = 2*seq->nrow*seq->ncol;
        xbnum = seq->nccd*seq->namps;
        for (i=0; i<xbnum; i++){
            im->pix[i]=(u_short *)malloc(xbsize);
        }
        printf("Image buffer size %d kB\n", (imsize>>10));
        printf("Extra buffer size %d kB\n", (xbsize>>10));
    }

    if(imbuf == NULL) {
        printf("Imbuf not allocated\n");
        return(ABORT);
    }

    /* Set up the exposure mode and start the exposure */
    seq->mode = NO_MODE;
    if (autoshutter==1) seq->mode |= AUTO_BEAM;
    if (autoflush) seq->mode |= AUTO_FLUSH;
    r = set_seqmode(seq->mode);
    r = exp_start();

    return(OK);
}

/*  read the ccd */
int
ccd_read(void)
{
    extern SEQ_MODE *seq;
    extern IMDESC *im;
    extern u_short *imbuf;
    extern int *dispbuf;
    extern int imsize, xbsize, xbnum;
    extern int autowipe, rdflag;
    struct timeval tp1,tp2;

    int i,j,k,l,m,r,nbytes,NLINE;
    int cnt,err;
    double diff;

    /* Get the current readout mode image size */
    NLINE = (seq->ncol*seq->nccd*seq->namps);
    seq->npix = (seq->nrow*NLINE);

    if(imbuf == NULL) {
        printf("Imbuf not allocated\n");
        return(ABORT);
    }

    gettimeofday(&tp1,NULL);
    err = 0;
    cnt = 1;
    m = 0;
    nbytes = 2*NLINE;

    if (rdflag == 0){
        (void)read_ccd();
        for (i=0; i<seq->nrow; i++){
            r = get_ccddata(imbuf, NLINE, &cnt);
            if (r!=nbytes){
                if(err==0)err=i;
                for (j=0; j<NLINE; j++) imbuf[j]=0x0000;
                r = 0;
            }
            m+=r;

            for (j=0,l=0; l<NLINE; l+=(seq->nccd*seq->namps),j++){
                for (k=0; k<(seq->nccd*seq->namps); k++){
                    im->pix[k][i*seq->ncol+j]=imbuf[l+k];
                }
            }
        }
        printf("ccd_read  retrys = %d     ",cnt);
        if (err>0)printf("Error in row = %d ",err);
        printf("\n");
    }

    if (rdflag == 1){
        for (i=0; i<seq->nrow; i++){
            (void)read_line();
            r = get_ccddata(imbuf, NLINE, &cnt);
            if (r!=nbytes){
                if(err==0)err=i;
                for (j=0; j<NLINE; j++) imbuf[j]=0x0000;
                r = 0;
            }
            m+=r;

            for (j=0,l=0; l<NLINE; l+=(seq->nccd*seq->namps), j++){
                for (k=0; k<(seq->nccd*seq->namps); k++){
                    im->pix[k][i*seq->ncol+j]=imbuf[l+k];
                }
            }
        }
        printf("ccd_read  retrys = %d     ",cnt);
        if (err>0)printf("Error in row = %d ",err);
        printf("\n");
    }

    if (rdflag == 2){
        (void)read_ccd();
        m = get_ccddata(imbuf, seq->nrow*NLINE, &cnt);
        printf("Total retrys= %d \n",cnt);

        for (j=0; j<seq->nrow*NLINE; j++){
            im->pix[0][j] = imbuf[j];
        }
        printf("\n");
    }

    gettimeofday(&tp2,NULL);
    diff = (double)(tp2.tv_sec-tp1.tv_sec)+(double)(tp2.tv_usec-tp1.tv_usec)/1.E6;

    /* Select the appropriate data buffer -- default is im->pix[0] */
    im->pixbuf = im->pix[0];
    
    if (seq->nccd > 1 && seq->namps == 1){
        im->pixbuf = im->pix[seq->opch];
    }

    // Rearrange the data order for two output amps per ccd
    if (seq->nccd == 1 && seq->namps == 2){
        for (i=0,j=0,k=(2*seq->ncol-1); i<(seq->nrow*seq->ncol); i++,j++,k--){
            imbuf[j]=im->pix[0][i];
            imbuf[k]=im->pix[1][i];
            if (i%(seq->ncol)==(seq->ncol-1)){
                j+=seq->ncol;
                k+=(3*seq->ncol);
            }
        }
        im->pixbuf = imbuf;
    }

#ifdef GCAM_TEST
    // Rearrange the data order for four output amps per ccd
    if (seq->nccd == 1 && seq->namps == 4){
        for (i=0, j=0, k=(2*seq->ncols-1), l=(2*seq->nrow-1)*(2*seq->ncol), 
            m=(2*seq->nrow*2*seq->ncol-1); i<(seq->nrow*seq->ncol); 
            i++,j++,k--,l++,m--){
            imbuf[j]=im->pix[0][i];
            imbuf[k]=im->pix[1][i];
            imbuf[l]=im->pix[2][i];
            imbuf[m]=im->pix[3][i];
            if (i%(seq->ncol)==(seq->ncol-1)){
                j+=seq->ncol;
                k+=(3*seq->ncol);
                l-=(3*seq->ncol);
                m-=(seq->ncol);
            }
        }
        im->pixbuf = imbuf;
    }
#endif // GCAM_TEST

    /* Finish up, reset autowipe if necessary */
    printf("Finished  r = %d kB     ",(int)(m>>10));
    printf("Elapsed time  t= %.1f sec\n",diff);

    if (autowipe) set_idlewipe();
    return(OK);
}

void
echo(void)
{
    static int sval=0;
    int rval;

    dsp_echo(sval, &rval);
    sval++;
}

/*===========================================================================*/
void
set_exptimer(void)
{
    extern SEQ_MODE *seq;
    extern int expflag;
    extern double exptime;
    extern int seqflag;
    int r;

    if (seqflag == SEQ_NONE){
        exptime = fp_arg();
    }

    if(exptime < MIN_EXPTIME){
        exptime = MIN_EXPTIME;
        printf("Invalid exposure time -- min %d\n",MIN_EXPTIME);
    }
    if (exptime > MAX_EXPTIME) {
        printf("Invalid exposure time -- max %d\n",MAX_EXPTIME);
        exptime = MAX_EXPTIME;
    }
    seq->seqtime = (int)(exptime * 1000.0);
    r = set_seqtime(seq->seqtime);
    ud_xtime();
    printf("Updated exptime (ms): %d,  Echo back: %d\n",seq->seqtime,r);
}

/* set the cycle time for exposure sequence */
void
set_cyctime(void)
{
    extern double cyctime;
    extern int seqflag;
    int r;

    if (seqflag == SEQ_NONE){
        cyctime = fp_arg();
    }

    if(cyctime < MIN_CYCTIME){
        cyctime = MIN_CYCTIME;
        printf("Invalid exposure time -- min %d\n",MIN_CYCTIME);
    }
    if (cyctime > MAX_CYCTIME) {
        printf("Invalid exposure time -- max %d\n",MAX_CYCTIME);
        cyctime = MAX_CYCTIME;
    }

    ud_cyctime();
    printf("Updated cyctime: %5.1f seconds\n",cyctime);

}

/* extend the exposure time */
void
tx_extend(void)
{
    extern double  exptime, endtime;
    double arg;

    if((arg = fp_arg()) < 0.0) return;

    exptime += arg;
    endtime += arg; 
    ud_xtime();
}

void
set_impath(void)
{
    extern char  impath[];
    char *nxttok;

    nxttok=nxttoken();
    if (nxttok == NULL) return;

    strcpy(impath, nxttok);
    ud_impath();

    printf("Set impath to %s\n", impath);
}

void
set_fprefix(void)
{
    extern char  fprefix[];
    char *nxttok;

    nxttok=nxttoken();
    if (nxttok == NULL) return;

    strcpy(fprefix, nxttok);
    printf("Set fprefix to %s\n", fprefix);
}

void
set_nfsave(void)
{
    extern int  nfsave;
    extern int seqflag;

    if (seqflag == SEQ_NONE){
        nfsave = (int)fp_arg();
    }

    if(nfsave < 0){
        nfsave = 0;
        printf("Invalid file number -- min %d\n",0);
    }
    if (nfsave > 999) {
        printf("Invalid file number-- max %d\n",999);
        nfsave = 999;
    }

    ud_nfsave();
    printf("Updated file number: %d\n",nfsave);
}

void
set_loop_num(void)
{
    extern int loop_num;
    extern int seqflag;

    if (seqflag == SEQ_NONE){
        loop_num = (int)fp_arg();
    }

    if(loop_num < 1){
        loop_num = 1;
        printf("Invalid loop number -- min %d\n",1);
    }
    if (loop_num > 999) {
        printf("Invalid loop number -- max %d\n",999);
        loop_num = 999;
    }

    ud_loop_num(loop_num);
    printf("Updated loop number: %d\n",loop_num);
}

void
set_cycle(void)
{
    extern int cycleflag;
    char *nxttok;

    nxttok=nxttoken();
    if (nxttok == NULL) return;

    if(strncmp(nxttok,"off",3) == 0) {
        cycleflag = CYCLE_OFF;
        ud_cycle(CYCLE_OFF);
    }
    else if(strncmp(nxttok,"on",2) == 0){
        cycleflag = CYCLE_ON;
        ud_cycle(CYCLE_ON);
    }
}


/*===========================================================================*/
void
set_autotemp(void)
{
    extern int autotemp;
    char *nxttok;

    nxttok=nxttoken();
    if (nxttok == NULL) return;

    if(strncmp(nxttok,"off",3) == 0) {
        autotemp = 0;
    }
    else if(strncmp(nxttok,"on",2) == 0){
        autotemp = 1;
    }
    ud_autotemp(autotemp);
}

void
set_autowipe(void)
{
    extern int autowipe, wipeflag;
    char *nxttok;

    nxttok=nxttoken();
    if (nxttok == NULL) return;

    if(strncmp(nxttok,"off",3) == 0) {
        autowipe = 0;
    }
    else if(strncmp(nxttok,"on",2) == 0){
        autowipe = 1;
    }
    ud_autowipe(autowipe);
    wipeflag = SET_WIPE;
}

void
set_autoflush(void)
{
    extern int autoflush;
    char *nxttok;
    int r;

    nxttok=nxttoken();
    if (nxttok == NULL){
        r = flush_ccd();
        if (r > 0) printf("Detector flushed \n");
    }

    if(strncmp(nxttok,"off",3) == 0) {
        autoflush = 0;
    }
    else if(strncmp(nxttok,"on",2) == 0){
        autoflush = 1;
    }
    ud_autoflush(autoflush);
}

void
set_autosave(void)
{
    extern int autosave;
    char *nxttok;
    int r;

    nxttok=nxttoken();
    if (nxttok == NULL)return;

    if(strncmp(nxttok,"off",3) == 0) {
        autosave = 0;
    }
    else if(strncmp(nxttok,"on",2) == 0){
        autosave = 1;
    }
    ud_autosave(autosave);
}

void
set_rdflag(void)
{
    extern int rdflag;
    char *nxttok;
    int r;

    nxttok=nxttoken();
    if (nxttok == NULL)return;

    if(strncmp(nxttok,"off",3) == 0) {
        rdflag = 0;
        printf("Set rdflag to 0x0\n");
    }
    else if(strncmp(nxttok,"on",2) == 0){
        rdflag = 1;
        printf("Set rdflag to 0x1\n");
    }
//    ud_rdflag(rdflag);
}

void
set_autoshutter(void)
{
    extern int autoshutter;
    char *nxttok;

    if((nxttok=nxttoken()) != NULL) {
        if(strncmp(nxttok,"auto",3) == 0) {
            ud_autoshutter(ASSERTED);
            autoshutter = ASSERTED;
            printf("Auto shutter enabled\n");
        }
        else if(strncmp(nxttok,"off",3) == 0){
            ud_autoshutter(!ASSERTED);
            autoshutter = !ASSERTED;
            printf("Auto shutter disabled\n");
        }
        else if(strncmp(nxttok,"open",3) == 0){
            ud_autoshutter(!ASSERTED);
            autoshutter = !ASSERTED;
            set_shutter(OPEN_SHUTTER);
        }
        else if(strncmp(nxttok,"close",3) == 0){
            ud_autoshutter(!ASSERTED);
            autoshutter = !ASSERTED;
            set_shutter(CLOSE_SHUTTER);
        }
        else if(strncmp(nxttok,"special",3) == 0){
            ud_autoshutter(!ASSERTED);
            autoshutter = SPECIAL;
            set_shutter(CLOSE_SHUTTER);
        }
        else {
            printf("Shutter mode not recognized (auto|off|open|close)\n");
        }
    }
}

/*
void
shutter(int arg)
{
    if(arg== OPEN_SHUTTER) {
        set_shutter(1);
    }
    else if(arg== CLOSE_SHUTTER){
        set_shutter(0);
    }
}
*/

/*========== Hardhat and housekeeping operations ===========================*/
void
hh_set_nrows(void)
{
    extern SEQ_MODE  *seq;
    extern int dd_state;
    extern int seqflag;
    int r;

    if (seqflag == SEQ_NONE){
        seq->nrow = (int)fp_arg();
    }
    if(seq->nrow < MIN_NROW){
        seq->nrow = MIN_NROW;
        printf("Invalid row spec -- min %d\n",MIN_NROW);
    }
    if (seq->nrow > MAX_NROW) {
        printf("Invalid row spec -- max %d\n",MAX_NROW);
        seq->nrow = MAX_NROW;
    }
    r = set_nrow(seq->nrow);
    ud_nrows(seq->nrow);
    printf("Updated nrows: %d,  Echo back: %d\n",seq->nrow,r);
    if (dd_state) dderase();
}

void
hh_set_ncols(void)
{
    extern SEQ_MODE  *seq;
    extern int dd_state;
    extern int seqflag;
    int r;

    if (seqflag == SEQ_NONE){
        seq->ncol = (int)fp_arg();
    }
    if(seq->ncol < MIN_NCOL){
        printf("Invalid ncol spec -- min %d\n",MIN_NCOL);
        seq->ncol = MIN_NCOL;
    }
    if (seq->ncol > MAX_NCOL) {
        printf("Invalid ncol spec -- max %d\n", MAX_NCOL);
        seq->ncol = MAX_NCOL;
    }
    r = set_ncol(seq->ncol);
    ud_ncols(seq->ncol);
    printf("Updated ncols: %d,  Echo back: %d\n", seq->ncol,r);
    if (dd_state) dderase();
}

void
hh_set_hbin(void)
{
    extern SEQ_MODE  *seq;
    extern int dd_state;
    extern int seqflag;
    int r;

    if (seqflag == SEQ_NONE){
        seq->hbin = (int)fp_arg();
    }
    if(seq->hbin < MIN_HBIN){
        printf("Invalid hbin spec -- min %d\n",MIN_HBIN);
        seq->hbin = MIN_HBIN;
    }
    if (seq->hbin > MAX_HBIN) {
        printf("Invalid hbin spec -- max %d\n", MAX_HBIN);
        seq->hbin = MAX_HBIN;
    }
    r = set_hbin(seq->hbin);
    ud_hbin(seq->hbin);
    printf("Updated hbin: %d,  Echo back: %d\n", seq->hbin,r);
    if (dd_state) dderase();
}

void
hh_set_vbin(void)
{
    extern SEQ_MODE  *seq;
    extern int dd_state;
    extern int seqflag;
    int r;

    if (seqflag == SEQ_NONE){
        seq->vbin = (int)fp_arg();
    }
    if(seq->vbin < MIN_VBIN){
        printf("Invalid vbin spec -- min %d\n",MIN_VBIN);
        seq->vbin = MIN_VBIN;
    }
    if (seq->vbin > MAX_VBIN) {
        printf("Invalid vbin spec -- max %d\n", MAX_VBIN);
        seq->vbin = MAX_VBIN;
    }
    r = set_vbin(seq->vbin);
    ud_vbin(seq->vbin);
    printf("Updated vbin: %d,  Echo back: %d\n", seq->vbin,r);
    if (dd_state) dderase();
}

void
hh_set_namps(void)
{
    extern SEQ_MODE *seq;
    extern int seqflag;
    extern int xbsize, xbnum;
    int r, namps;

    if (seqflag == SEQ_NONE){
        seq->namps = (int)fp_arg();
    }
    if(seq->namps < MIN_NAMPS){
        printf("Invalid namps spec -- min %d\n",MIN_NAMPS);
        seq->namps = MIN_NAMPS;
    }
    if (seq->namps > MAX_NAMPS) {
        printf("Invalid namps spec -- max %d\n", MAX_NAMPS);
        seq->namps = MAX_NAMPS;
    }
    r = set_namps(seq->namps);
    ud_namps(seq->namps);
    printf("Updated namps: %d,  Echo back: %d\n", seq->namps,r);
    xbnum = 0;    // xbsize = 0;
}

void
hh_set_nccd(void)
{
    extern SEQ_MODE *seq;
    extern int seqflag;
    extern int xbsize, xbnum;
    int r, nccd;

    if (seqflag == SEQ_NONE){
        seq->nccd = (int)fp_arg();
    }
    if(seq->nccd < MIN_NCCD){
        printf("Invalid nccd spec -- min %d\n",MIN_NCCD);
        seq->nccd = MIN_NCCD;
    }
    if (seq->nccd > MAX_NCCD) {
        printf("Invalid nccd spec -- max %d\n", MAX_NCCD);
        seq->nccd = MAX_NCCD;
    }
    r = set_nccd(seq->nccd);
    ud_nccd(seq->nccd);
    printf("Updated nccd: %d,  Echo back: %d\n", seq->nccd,r);
    xbnum = 0;   // xbsize = 0;
}

void
hh_set_nch(void)
{
}

void
hh_set_gain(void)
{
    extern SEQ_MODE  *seq;
    extern int seqflag;
    int r;
    char *nxttok;

    if (seqflag == SEQ_NONE){
        nxttok=nxttoken();
        if (nxttok == NULL) return;

        if(strncmp(nxttok,"low",3) == 0) {
            seq->gain = 0;
        }
        else if(strncmp(nxttok,"high",4) == 0){
            seq->gain = 1;
        }
    }
    if(seq->gain < MIN_GAIN){
        printf("Invalid gain spec -- min %d\n",MIN_GAIN);
        seq->gain = MIN_GAIN;
    }
    if (seq->gain > MAX_GAIN) {
        printf("Invalid gain spec -- max %d\n", MAX_GAIN);
        seq->gain = MAX_GAIN;
    }
    r = set_gain(seq->gain);
    ud_gain(seq->gain);
    printf("Updated gain: %d,  Echo back: %d\n", seq->gain,r);
}

void
hh_set_usec(void)
{
    extern SEQ_MODE  *seq;
    extern int seqflag;
    int r;

    if (seqflag == SEQ_NONE){
        seq->usec = (int)fp_arg();
    }
    if(seq->usec < MIN_USEC){
        printf("Invalid usec spec -- min %d\n",MIN_USEC);
        seq->usec = MIN_USEC;
    }
    if (seq->usec > MAX_USEC) {
        printf("Invalid usec spec -- max %d\n", MAX_USEC);
        seq->usec = MAX_USEC;
    }
    r = set_usec(seq->usec);
    ud_usec(seq->usec);
    printf("Updated usec: %d,  Echo back: %d\n", seq->usec,r);
}


void
hh_set_opch(void)          /* Set serial reg output amp */
{
    extern IMDESC *im;
    extern SEQ_MODE  *seq;
    extern int seqflag, expflag, showflag;
    int r;
    char *nxttok;

#ifdef GCAM
    if (seqflag == SEQ_NONE){
        nxttok=nxttoken();
        if (nxttok == NULL) return;

        if(strncmp(nxttok,"A",1) == 0) {
            seq->opch = 0x1;
        }
        else if(strncmp(nxttok,"B",1) == 0){
            seq->opch = 0x2;
        }
        else if(strncmp(nxttok,"Both",4) == 0){
            seq->opch = 0x3;
        }
        else if(strncmp(nxttok,"Four",4) == 0){
            seq->opch = 0x4;
        }
    }
#endif // GCAM
#ifdef BASE
    if (seqflag == SEQ_NONE){
        nxttok=nxttoken();
        if (nxttok == NULL) return;

        if(strncmp(nxttok,"A",1) == 0) {
            seq->opch = 0x0;
        }
        else if(strncmp(nxttok,"B",1) == 0){
            seq->opch = 0x1;
        }
        else if(strncmp(nxttok,"C",1) == 0){
            seq->opch = 0x2;
        }
        else if(strncmp(nxttok,"D",1) == 0){
            seq->opch = 0x3;
        }
        else if(strncmp(nxttok,"E",1) == 0){
            seq->opch = 0x4;
        }
        else if(strncmp(nxttok,"F",1) == 0){
            seq->opch = 0x5;
        }
        else if(strncmp(nxttok,"G",1) == 0){
            seq->opch = 0x6;
        }
        else if(strncmp(nxttok,"H",1) == 0){
            seq->opch = 0x7;
        }
    }
#endif // BASE

    if(seq->opch < MIN_OPCH){
        printf("Invalid output channel -- min %d\n",MIN_OPCH);
        seq->opch = MIN_OPCH;
    }
    if (seq->opch > MAX_OPCH) {
        printf("Invalid output channel -- max %d\n", MAX_OPCH);
        seq->opch = MAX_OPCH;
    }
    r = set_opch(seq->opch);
    ud_opch(seq->opch);
    printf("Updated output channel: %d,  Echo back: %d\n", seq->opch,r);

#ifdef GCAM
    if (seq->opch==1)seq->namps=1;
    if (seq->opch==2)seq->namps=1;
    if (seq->opch==3)seq->namps=2;
    if (seq->opch==4)seq->namps=4;
    r = set_namps(seq->namps);
#endif // GCAM

#ifdef BASE
    /* Select the appropriate data buffer */
    if (im->nccd == 1){
        im->pixbuf = im->pix[0];
    }
    else if (seq->opch <= im->nccd){
        im->pixbuf = im->pix[seq->opch];
        printf("Showing new output\n");
        expflag = SHOWNEW;
    }
#endif // BASE
}

void
hh_set_sclks(void)
{
    extern SEQ_MODE *seq;
    extern int seqflag;
    int r;
    char *nxttok;

    if (seqflag == SEQ_NONE){
        nxttok=nxttoken();
        if (nxttok == NULL) return;

        if(strncmp(nxttok,"left",4) == 0) {
            seq->sclks = 0x1;
        }
        else if(strncmp(nxttok,"right",5) == 0){
            seq->sclks = 0x2;
        }
    }

#ifdef BASE
    r = set_sclks(seq->sclks);
    printf("Updated sclks: %d,  Echo back: %d\n", seq->sclks,r);
#endif // BASE
#ifdef GCAM
    printf("Update sclks -- not implemented yet\n");
#endif // GCAM
}

void
hh_set_pid(void)
{
    extern SEQ_MODE *seq;
    extern int seqflag, pidflag;
    int j,r;
    char *nxttok;

    if (seqflag == SEQ_NONE){
        nxttok=nxttoken();
        if (nxttok == NULL) return;

        if(strncmp(nxttok,"on",2) == 0) {
            seq->pid = 0x1;
            pidflag = ASSERTED;
        }
        else if(strncmp(nxttok,"off",3) == 0){
            seq->pid = 0x0;
            pidflag = !ASSERTED;
        }
    }

#ifdef BASE
    r = set_pid(seq->pid);
    ud_autopid(pidflag);
    printf("Updated temp pid: %d,  Echo back: %d\n", seq->pid,r);

    if (seq->pid==0){
        for (j=0; j<8; j++)seq->heat[j]=0.0;
        seqflag = SET_HEAT;
    }
#endif // BASE
    printf("Update pid asserted\n");
}

void
hh_exp_status(void)
{
#ifdef BASE
    extern SEQ_MODE *seq;
    extern int seqflag;
    int a,b,r;

    r = exp_status(&a,&b);
    printf("Exp  State= %06x  Timer= %06x \n", a,b);
#endif // BASE
}

void
hh_set_vdacs(void)
{
    extern SEQ_MODE *seq;
    extern int hhflag;
    int j, k, r;

#ifdef BASE
    base_set_vdac(seq->vbias,seq->vclk,seq->opch);
#endif // BASE

#ifdef GCAM
    gcam_set_vdac(seq->vbias,seq->vclk);
#endif // GCAM
    printf("Updated ccd voltages\n");
}

void
hh_get_vdacs(void)
{
    extern SEQ_MODE *seq;
    extern int hhflag;
    int j, k, r;

#ifdef BASE
    base_get_vdac(seq->vbias,seq->vclk,seq->opch);
#endif // BASE

#ifdef GCAM
    gcam_get_vdac(seq->vbias,seq->vclk);
#endif // GCAM

    ud_hhpanel();
}

void
hh_set_ccdtemp(void)
{
#ifdef BASE
    extern SEQ_MODE *seq;
    extern int verbose;
    static int index=0;
    int j, k, r, arg;

    base_set_tset(seq->tset);
    ud_temp();

    if (verbose==1){
        printf("Tset=  %5.1f %5.1f %5.1f %5.1f",
            seq->tset[0],seq->tset[1],seq->tset[2],seq->tset[3]);
        printf(" %5.1f %5.1f %5.1f %5.1f\n",
            seq->tset[4],seq->tset[5],seq->tset[6],seq->tset[7]);
    }
#endif // BASE
    printf("Updated ccd temperature setpoint(s)\n");
}

void
hh_get_ccdtemp(void)
{
    extern SEQ_MODE *seq;
    extern int hhflag, pidflag, autotemp, verbose;
    int j, k, r, arg;

#ifdef GCAM
    r = gcam_get_temp(seq->temp);
    if (r!=0) {
        autotemp=0;
        ud_autotemp(autotemp);
    }
    ud_temp();

    if (verbose==1){
        printf("Temp= %5.1f \n", seq->temp[0]);
    }
#endif // GCAM

#ifdef BASE
    r = base_get_temp(seq->temp);
    if (r!=0) {
        autotemp=0;
        ud_autotemp(autotemp);
    }

    r = base_get_tset(seq->tset);
    ud_temp();

    if (verbose==1){
        printf("Temp= %5.1f %5.1f %5.1f %5.1f",
            seq->temp[0],seq->temp[1],seq->temp[2],seq->temp[3]);
        printf(" %5.1f %5.1f %5.1f %5.1f\n",
            seq->temp[4],seq->temp[5],seq->temp[6],seq->temp[7]);
    }

    if (pidflag==1){
        base_get_heat(seq->heat);
        ud_heat();
//      base_pid(seq->tset, seq->temp, seq->heat, seq->opch);     // debug-only
    }
    if (pidflag==1 && verbose==1){
        printf("Heat= %5.1f %5.1f %5.1f %5.1f",
            seq->heat[0],seq->heat[1],seq->heat[2],seq->heat[3]);
        printf(" %5.1f %5.1f %5.1f %5.1f\n",
            seq->heat[4],seq->heat[5],seq->heat[6],seq->heat[7]);
    }
#endif // BASE
}

void
hh_set_heat(void)           // Set heat or TEC value
{
    extern SEQ_MODE *seq;
    extern int verbose, hhflag, seqflag;
    int r;
    double val;

#ifdef GCAM
    if (hhflag == IDLE){
        val = (double)fp_arg();
    }
    if(val < MIN_TEC){
        val = MIN_TEC;
        printf("Invalid TEC spec -- min %5.1f\n",MIN_TEC);
    }
    if (val > MAX_TEC) {
        printf("Invalid TEC spec -- max %5.1f\n",MAX_TEC);
        val = MAX_TEC;
    }

    seq->tec[0] = val;
    seq->vbias[2] = val;
    r = gcam_set_tec(seq->tec);
    printf("TEC=  %6.1f \n", seq->tec[0]);
    ud_heat();
#endif // GCAM

#ifdef BASE
    base_set_heat(seq->heat);
    ud_heat();

    if (verbose==1){
        printf("HEAT=  %5.1f %5.1f %5.1f %5.1f",
            seq->heat[0],seq->heat[1],seq->heat[2],seq->heat[3]);
        printf(" %5.1f %5.1f %5.1f %5.1f\n",
            seq->heat[4],seq->heat[5],seq->heat[6],seq->heat[7]);
    }

    printf("Updated heater setpoints\n");

//  Turn off PID if heat[] set manually
//    seq->pid = 0x0;
//    pidflag = !ASSERTED;
//    r = set_pid(seq->pid);
//    ud_autopid(pidflag);
//    printf("Updated temp pid: %d,  Echo back: %d\n", seq->pid,r);

#endif // BASE
}

void
hh_get_heat(void)
{
    extern SEQ_MODE *seq;
    extern int verbose, hhflag;
    int r;
    double val;

#ifdef GCAM
//    gcam_get_tec(seq->tec);
//    ud_heat();
//    if (verbose==1){
//        printf("TEC=  %5.1f mA\n", 1000.0*seq->tec);
//    }
#endif // GCAM

#ifdef BASE
    base_get_heat(seq->heat);
    ud_heat();

    if (verbose==1){
        printf("Heat=  %5.1f %5.1f %5.1f %5.1f",
            seq->heat[0],seq->heat[1],seq->heat[2],seq->heat[3]);
        printf(" %5.1f %5.1f %5.1f %5.1f\n",
            seq->heat[4],seq->heat[5],seq->heat[6],seq->heat[7]);
    }
#endif // BASE
}

/*===========================================================================*/
void
show_image(void)
{
    extern int dd_state;
    extern IMDESC *im;

    /* show current image -- pipe it to imtool */
    if(dd_state) 
        ddwriteimage(im);
    else 
        printf("Display device is not open\n");
}

void
save_fitsimage(void)
{
    extern IMDESC *im;
    extern char impath[], fname[];
    extern char *suffix[];
    extern char fprefix[];
    extern int expflag;
    extern int nfsave;
    char *name, bname[80];
    int i,r;

    if (expflag == IDLE){
        name = nxttoken();
        if(name != NULL) {
            sprintf(bname, "%s%s", impath, name);
        }
        else {
            printf("need file name, try again\n");
            return;
        }
    }

    if (expflag == FILESAVE){
        sprintf(bname,"%s%s%3.3d",impath,fprefix,nfsave);
    }

    if (im->nccd == 1){
        sprintf(fname,"%s.fits",bname);
        r=ccdwrfits(im, fname);
        printf("Saved: %s\n", fname);
    }
    else {
        for (i=0; i<im->nccd; i++){
            sprintf(fname,"%s%s.fits",bname, suffix[i]);
            im->pixbuf = im->pix[i];
            r=ccdwrfits(im, fname);
            printf("Autosaved: %s\n", fname);
        }
    }
}

/*  open command file & parse  */
void
cx_scriptrun(void)
{
    char  *fname, fpath[80];

    fname = nxttoken();
    scriptrun(fname);

}

void 
get_cmdline(void)
{
    extern char cmd1[], cmd2[], cmd3[], cmd4[]; 
    int cmdnum, r;
    char *nxttok, *cmdbuf;

    if((cmdnum = (int)fp_arg()) <= 0) { 
        printf("  %s\n", cmd1);
        printf("  %s\n", cmd2);
        printf("  %s\n", cmd3);
        printf("  %s\n", cmd4); 
        return;
    }

    switch (cmdnum) {
    case 1 : 
        cmdbuf = cmd1;  
        break;
    case 2 : 
        cmdbuf = cmd2;  
        break;
    case 3 : 
        cmdbuf = cmd3;  
        break;
    case 4 : 
        cmdbuf = cmd4;  
        break;
    default : 
        printf("command not defined: cmd%d\n", cmdnum);
        return;
    }

    *cmdbuf = 00;
    while( (nxttok=nxttoken()) != NULL) {
        strcat(cmdbuf, nxttok);
        strcat(cmdbuf, " ");
    }
}

/*========================================================================*
 *     Display device and plot functions 
 *========================================================================*/
void
set_ddconfig(void)
{

}

void
set_xrange(void)
{
    int min, max, r;

    if((min = (int)fp_arg()) < -1) return;
    if((max = (int)fp_arg()) < -1) return;

    if (max > MAX_NCOL) max = MAX_NCOL;

    pf_setxrange(min, max);
    printf("x plot range min->max : %d->%d\n", min, max);
    pf_replot();
}

void
set_yrange(void)
{
    int min, max, r;

    if((min = (int)fp_arg()) < -1) return;
    if((max = (int)fp_arg()) < -1) return;

    if (max > USHORTMAX) max = USHORTMAX;

    pf_setyrange(min, max);
    printf("y plot range min->max : %d->%d\n", min, max);
    pf_replot();
}

void
set_zrange(void)
{
    int min, max, r;

    if((min = (int)fp_arg()) < -1) return;
    if((max = (int)fp_arg()) < -1) return;

    if (max > USHORTMAX) max=USHORTMAX;

//    ddsetzrange(min, max);
    printf("display z min->max : %d->%d\n", min, max); 
}

void
set_zmin(void)
{
    int min, r;

    if((min = (int)fp_arg()) < -1) return;

    if (min > USHORTMAX) min=0;

//    ddsetzmin(min);
    printf("Set display z min: %d\n", min);
    show_image();
}

void
set_zmax(void)
{
    int max, r;

    if((max = (int)fp_arg()) < -1) return;

    if (max > USHORTMAX) max=USHORTMAX;

//    ddsetzmax(max);
    printf("Set display z max: %d\n", max);
    show_image();
}

void
set_zspan(void)
{
    int span, r;

    if((span = (int)fp_arg()) < -1) return;

    if (span > USHORTMAX) span=USHORTMAX;

//    ddsetzspan(span);
    printf("Set display z span: %d\n", span);
    show_image();
}

void
plot_col(void)
{
    extern IMDESC  *im;
    int col, r; 

    if((col = (int)fp_arg()) < 0) return; 
    pf_plotcol(im, col, 1); 
}

void
plot_row(void)
{
    extern IMDESC  *im;
    int row, r; 
    int j;

    if((row = (int)fp_arg()) < 0) return; 

    pf_plotrow(im, row, 1); 
}

void
plot_coladd(void)
{
    extern IMDESC  *im;
    int col, r; 

    if((col = (int)fp_arg()) < 0) return; 
    pf_plotcol(im, col, 0); 
}

void
plot_rowadd(void)
{
    extern IMDESC  *im;
    int row, r; 
    
    if((row = (int)fp_arg()) < 0) return; 
    pf_plotrow(im, row, 0); 
}


/*==================== End Of File ======================================== */
