/*============================================================================*/
/*     Header file for dsp interaction routines                               */
/*============================================================================*/

/* Define command codes to perform the following actions: */
#define DSP_ABORT       0x00    /* Abort DSP operations & wait */
#define DSP_ECHO        0x01    /* Echo word back to host */
#define DSP_GET         0x02    /* Read control parameters from DSP */
#define DSP_PUT         0x03    /* Load control parameters into DSP */
#define FLUSH           0x04    /* Flush the CCD */
#define BEAM_ON         0x05    /* Open the shutter */
#define EXPCCD          0x06    /* CCD exposure timer */
#define EXP_START       0x06    /* Start the exposure timer */
#define BEAM_OFF        0x07    /* Close the shutter */
#define READCCD         0x08    /* Read the CCD and send the data */
#define READOUT         0x08    /* Read the detector and send the data */
#define IDLE_WIPE       0x09    /* Auto-wipe the CCD while idle */

#define HH_DACS         0x0A    /* Update DAC values */
#define HH_TEMP         0x0B    /* Get temperature value */
#define HH_TEC          0x0C    /* Set TEC value */
#define HH_OTHER        0x0D    /* Housekeeping */

#define EXP_PAUSE       0x0E
#define EXP_RESUME      0x0F
#define SHUFFLE_F       0x10
#define SHUFFLE_R       0x11
#define EXP_STATUS      0x12
#define READLINE        0x13
#define FLUSHLINE       0x14
#define FRAME_XFER      0x15
#define TDI             0x16    /* reserved, but not used yet */
#define DSP_DUMMY       0x17

#define REBOOT          0x18

/* Define bootloader commands */
#define  BOOT_SN        0xFF00      /* Get DS2401 silicon serial number */
#define  BOOT_LOAD      0xFE01      /* Upload code into dsp */

/* Define options for the exposure mode (opflags) */
#define NO_SEND         0x0001      /* Data send-no-send flag */
#define AUTO_FLUSH      0x0002      /* Auto flush before exposure */
#define AUTO_BEAM       0x0004      /* Auto open-close shutter */
#define AUTO_EXPOSE     0x0008      /* Auto exposure timer */
#define AUTO_READ       0x0010      /* Auto readout after exposure */
#define AUTO_WIPE       0x0020      /* Auto wipe when idle */
#define AUTO_SEND       0x0040      /* Auto send the data */
#define AUTO_RESET      0x0080      /* Auto reset/read (H2RG) */
#define NO_CLK          0x0100      /* Skip clock cycle */

#define TDI_MODE        0x2000      /* TDI readout mode */
#define WIN_MODE        0x4000      /* Window mode - full or subraster */
#define EXP_IN_PROG     0x8000      /* Exposure in progress flag */

#ifdef DSPDVR_C
int dsp_open(void);
void dsp_close(void);
void dsp_reset(void);
int dsp_clear_fifo(void);
int dsp_init(char *srec_file);
int dsp_initII(char *srec_file);
int dsp_cmd(int cmd);
int dsp_write16(int value);
int dsp_read16(int *value);
int dsp_write24(int value);
int dsp_read24(int *value);
int dsp_echo(int sval, int *rval);
int dsp_put(int item, int value);
int dsp_get(int item, int* value);
int dsp_abort(void);
int set_seqmode(int mode);
int set_shutter(int arg);
int exp_start(void);
int exp_pause(void);
int exp_resume(void);
int exp_status(int *estate, int *etimer);
//int exp_seqstatus(int *estate, int *etimer1, int *etimer2, int *loopcnt);
int read_ccd(void);
int read_line(void);
int get_ccddata(u_short pixbuf[], int npix, int *cnt);
int flush_ccd(void);
int flush_line(void);
int shuffle(int dir);
int set_idlewipe(void);
int set_nrow(int nrow);
int get_nrow(void);
int set_ncol(int ncol);
int get_ncol(void);
int set_nft(int nft);
int get_nft(void);
int set_nflush(int nflush);
int get_nflush(void);
int set_nccd(int nccd);
int get_nccd(void);
int set_namps(int namps);
int get_namps(void);
int set_nch(int nch);
int get_nch(void);
int set_npix(int npix);
int get_npix(void);
int set_vbin(int vbin);
int get_vbin(void);
int set_hbin(int hbin);
int get_hbin(void);
int set_vskip(int vskip);
int get_vskip(void);
int set_hskip(int hskip);
int get_hskip(void);
int set_vsub(int vsub);
int get_vsub(void);
int set_hsub(int hsub);
int get_hsub(void);
//int set_nexp(int nexp);
//int get_nexp(void);
int set_nshuffle(int nshuffle);
int get_nshuffle(void);
int set_seqtime(int exp_time);
int get_seqtime(void);
//int set_exptime(int exp_time);
//int get_exptime(void);
//int set_exptime1(int exp_time1);
//int get_exptime1(void);
//int set_exptime2(int exp_time2);
//int get_exptime2(void);
int set_gain(int gain);
int get_gain(void);
int set_usec(int usec);
int get_usec(void);
int set_opch(int opch);
int get_opch(void);
int set_sclks(int sclks);
int get_sclks(void);
int set_pid(int pid);
int get_pid(void);
//int set_wm(int wm);
//int get_wm(void);
//int set_nreads(int nreads);
//int get_nreads(void);
int set_dacs(void);
int base_set_temp(double temp[]);
int base_get_temp(double temp[]);
int base_set_tset(double tset[]);
int base_get_tset(double tset[]);
int base_set_heat(double heat[]);
int base_get_heat(double heat[]);
int base_set_vdac(double vbias[], double vclk[],int ccd);
int base_get_vdac(double vbias[], double vclk[], int ccd);
int base_set_vbias(double vbias[], int ccd);
int base_get_vbias(double vbias[], int ccd);
int base_set_vclks(double vclk[], int ccd);
int base_get_vclks(double vclks[], int ccd);
int base_pid(double tset[], double temp[], double heat[], int ccd);
int gcam_set_tec(double tec[]);
int gcam_get_tec(double tec[]);
int gcam_get_temp(double temp[]);
int gcam_set_vdac(double vbias[], double vclk[]);
int gcam_get_vdac(double vbias[], double vclk[]);
int gcam_set_vbias(double vbias[]);
int gcam_get_vbias(double vbias[]);
int gcam_set_vclks(double vclk[]);
int gcam_get_vclks(double vclks[]);
//int h2rg_get_temp(double temp[]);
//int h2rg_set_tset(double tset[]);
//int h2rg_get_tset(double tset[]);
//int h2rg_set_heat(double heat[]);
//int h2rg_get_heat(double heat[]);
//int h2rg_set_vdac(double vbias[]);
//int h2rg_get_vdac(double vbias[]);
//int h2rg_set_vbias(double vbias[]);
//int h2rg_get_vbias(double vbias[]);
//int h2rg_get_tstamps(int tstamp[]);
#endif

#ifndef DSPDVR_C
extern int dsp_open(void);
extern int dsp_close(void);
extern int dsp_reset(void);
extern int dsp_clear_fifo(void);
extern int dsp_init(char *srec_file);
extern int dsp_initII(char *srec_file);
extern int dsp_cmd(int cmd);
extern int dsp_write16(int value);
extern int dsp_read16(int *value);
extern int dsp_write24(int value);
extern int dsp_read24(int *value);
extern int dsp_echo(int sval, int *rval);
extern int dsp_put(int item, int value);
extern int dsp_get(int item, int* value);
extern int dsp_abort(void);
extern int set_seqmode(int mode);
extern int set_shutter(int arg);
extern int exp_start(void);
extern int exp_pause(void);
extern int exp_resume(void);
extern int exp_status(int *estate, int *etimer);
//extern int exp_seqstatus(int *estate, int *etimer1, int *etimer2, int *loopcnt);
extern int read_ccd(void);
extern int read_line(void);
extern int get_ccddata(u_short pixbuf[], int npix, int *cnt);
extern int flush_ccd(void);
extern int flush_line(void);
extern int shuffle(int dir);
extern int set_idlewipe(void);
extern int set_nrow(int nrow);
extern int get_nrow(void);
extern int set_ncol(int ncol);
extern int get_ncol(void);
extern int set_nft(int nft);
extern int get_nft(void);
extern int set_nflush(int nflush);
extern int get_nflush(void);
extern int set_nccd(int nccd);
extern int get_nccd(void);
extern int set_namps(int namps);
extern int get_namps(void);
extern int set_nch(int nch);
extern int get_nch(void);
extern int set_npix(int npix);
extern int get_npix(void);
extern int set_vbin(int vbin);
extern int get_vbin(void);
extern int set_hbin(int hbin);
extern int get_hbin(void);
extern int set_vskip(int vskip);
extern int get_vskip(void);
extern int set_hskip(int hskip);
extern int get_hskip(void);
extern int set_vsub(int vsub);
extern int get_vsub(void);
extern int set_hsub(int hsub);
extern int get_hsub(void);
//extern int set_nexp(int nexp);
//extern int get_nexp(void);
extern int set_nshuffle(int nshuffle);
extern int get_nshuffle(void);
extern int set_seqtime(int exp_time);
extern int get_seqtime(void);
//extern int set_exptime(int exp_time);
//extern int get_exptime(void);
//extern int set_exptime1(int exp_time1);
//extern int get_exptime1(void);
//extern int set_exptime2(int exp_time2);
//extern int get_exptime2(void);
extern int set_gain(int gain);
extern int get_gain(void);
extern int set_usec(int usec);
extern int get_usec(void);
extern int set_opch(int opch);
extern int get_opch(void);
extern int set_sclks(int sclks);
extern int get_sclks(void);
extern int set_pid(int pid);
extern int get_pid(void);
//extern int set_wm(int wm);
//extern int get_wm(void);
//extern int set_nreads(int nreads);
//extern int get_nreads(void);
extern int set_dacs(void);
extern int base_set_temp(double temp[]);
extern int base_get_temp(double temp[]);
extern int base_set_tset(double tset[]);
extern int base_get_tset(double tset[]);
extern int base_set_heat(double heat[]);
extern int base_get_heat(double heat[]);
extern int base_set_vdac(double vbias[], double vclk[], int ccd);
extern int base_get_vdac(double vbias[], double vclk[], int ccd);
extern int base_set_vbias(double vbias[], int ccd);
extern int base_get_vbias(double vbias[], int ccd);
extern int base_set_vclks(double vclk[], int ccd);
extern int base_get_vclks(double vclks[], int ccd);
extern int base_pid(double tset[], double temp[], double heat[], int ccd);
extern int gcam_set_tec(double tec[]);
extern int gcam_get_tec(double tec[]);
extern int gcam_get_temp(double temp[]);
extern int gcam_set_vdac(double vbias[], double vclk[]);
extern int gcam_get_vdac(double vbias[], double vclk[]);
extern int gcam_set_vbias(double vbias[]);
extern int gcam_get_vbias(double vbias[]);
extern int gcam_set_vclks(double vclk[]);
extern int gcam_get_vclks(double vclks[]);
//extern int h2rg_get_temp(double temp[]);
//extern int h2rg_set_tset(double tset[]);
//extern int h2rg_get_tset(double tset[]);
//extern int h2rg_set_heat(double heat[]);
//extern int h2rg_get_heat(double heat[]);
//extern int h2rg_set_vdac(double vbias[]);
//extern int h2rg_get_vdac(double vbias[]);
//extern int h2rg_set_vbias(double vbias[]);
//extern int h2rg_get_vbias(double vbias[]);
//extern int h2rg_get_tstamps(int tstamp[]);
#endif
