/*============================================================================*/
/*     Header file for MOTXER application                                     */
/*============================================================================*/
#define  OK             0
#define  ABORT          -1

#ifndef  TRUE
#define  TRUE           1
#define  FALSE          0
#endif

#define  USHORTMAX          65535

#define  MAX_NROW           5000    /* number of rows in image */
#define  MIN_NROW           32
#define  MAX_NCOL           5000    /* number of columns in image */
#define  MIN_NCOL           32
#define  MAX_NFT            5000    /* number of frame transfer rows */
#define  MIN_NFT            32
#define  MAX_NFLUSH         5000    /* number of columns to flush */
#define  MIN_NFLUSH         32
#define  MAX_NCH            8       /* number of output/readout channels */
#define  MIN_NCH            1
#define  MAX_NCCD           8       /* number of detectors */
#define  MIN_NCCD           1
#define  MAX_NAMPS          4       /* number of output amps per chip */
#define  MIN_NAMPS          1
#define  MAX_VBIN           8       /* vertical (parallel) binning */
#define  MIN_VBIN           1
#define  MAX_HBIN           8       /* horizontal (serial) binning */
#define  MIN_HBIN           1
#define  MAX_VSKIP          4096    /* Offset to subraster */
#define  MIN_VSKIP          1
#define  MAX_HSKIP          2048    /* Offset to subraster */
#define  MIN_HSKIP          1
#define  MAX_VSUB           4100      /* Subraster size */
#define  MIN_VSUB           32
#define  MAX_HSUB           4100      /* Subraster size */
#define  MIN_HSUB           32

#define  MAX_SEQTIME        10*3600*1000    /* Sequencer time (1ms units) */
#define  MIN_SEQTIME        1
#define  MAX_NSHUFFLE       1       /* Nod/shuffle parameter */
#define  MIN_NSHUFFLE       0
#define  MAX_GAIN           1       /* Sig_proc gain */
#define  MIN_GAIN           0
#define  MAX_OPCH           7       /* Output channel number */
#define  MIN_OPCH           0
#define  MAX_USEC           2500    /* Sig_proc CDS sample time */
#define  MIN_USEC           10
#define  MAX_SCLKS          3       /* Serial clock direction */
#define  MIN_SCLKS          1
#define  MAX_PID            1       /* PID temp-control flag */
#define  MIN_PID            0

#define  MAX_NFSAVE         999     /* file number suffix */
#define  MIN_NFSAVE         0
#define  MAX_LOOPNUM        999     /* loop number */
#define  MIN_LOOPNUM        1

#define  MAX_EXPTIME        36000   /* Exposure time (s) */
#define  MIN_EXPTIME        0     
#define  MAX_CYCTIME        3600    /* Cycle time (s) */
#define  MIN_CYCTIME        0     

#define  MAX_TEC            3.0     /* TEC current (amps) */
#define  MIN_TEC            0.0
#define  MAX_TEMP           100.0   /* Temperature measurement (degC) */
#define  MIN_TEMP           -200.0
#define  MAX_HEAT           250.0   /* Heater current (mA) */
#define  MIN_HEAT           0.0
#define  MAX_TSET           100.0   /* Temperature set-point */
#define  MIN_TSET           -200.0

#define  DEFAULT_NROW       500
#define  DEFAULT_NCOL       500
#define  DEFAULT_NFT        500
#define  DEFAULT_NFLUSH     500
#define  DEFAULT_VSUB       40
#define  DEFAULT_HSUB       40

#define  DEFAULT_SEQTIME    1000
#define  DEFAULT_TIMEOUT    10      /* default link timeout */
#define  DEFAULT_IMSIZE     2*500*500  /* default image size (bytes) */

/* exposure state control flags used with expflg */
#define  IDLE               0x0000
#define  CCDFLUSH           0x0001  /* Flush the detector */
#define  EXPSTART           0x0002  /* start the exposure timer */
#define  EXPOSE             0x0004  /* exposure timer is running */
#define  EXPABORT           0x0008  /* abort exposure -- no readout */
#define  CCDREAD            0x0010  /* Read the detector */
#define  SHOWNEW            0x0020  /* Display the image */
#define  FILESAVE           0x0040  /* Save the file to disk */
#define  DELAYSTART         0x0080  /* start the delay timer */
#define  EXPDELAY           0x0100  /* delay timer is running */
#define  NEXTCYCLE          0x0200
#define  HHSYNC             0x0400  /* sync hard-hat values */

/* control flags settings */
#define  CYCLE_ON           0x01
#define  CYCLE_OFF          0x00

#define  OPEN_SHUTTER       0x01
#define  CLOSE_SHUTTER      0x00

#define  SPECIAL            0x02
#define  ASSERTED           0x01
#define  NOT_ASSERTED       0x00

#define  SEQ_NONE           0x00
#define  SET_VDAC           0x01
#define  GET_VDAC           0x02
#define  SET_HEAT           0x03
#define  GET_TEMP           0x04
#define  SET_WIPE           0x05
#define  SET_NROWS          0x06
#define  SET_NCOLS          0x07
#define  SET_HBIN           0x08
#define  SET_VBIN           0x09
#define  SET_GAIN           0x0A
#define  SET_OPCH           0x0B
#define  SET_XTIME          0x0C
#define  SET_CYCTIME        0x0D
#define  SET_LOOPNUM        0x0E
#define  SET_FSAVE          0x0F
#define  SET_USEC           0x10
#define  SET_NCCD           0x11
#define  SET_TEMP           0x12
#define  GET_HEAT           0x13
#define  SET_PID            0x14
#define  SET_SCLKS          0x15

/* Define options for the exposure mode (opflags) */
#define NO_MODE         0x0000
#define AUTO_FLUSH      0x0002
#define AUTO_BEAM       0x0004
#define AUTO_EXPOSE     0x0008
#define AUTO_READ       0x0010
#define AUTO_WIPE       0x0020
#define AUTO_SEND       0x0040

#ifdef GCAM
//#define  NBIAS              8
//#define  NCLKS              16
//#define  NTEMP              8
//#define  NHEAT              8
#endif // GCAM

#ifdef BASE
//#define  NBIAS              16
//#define  NCLKS              32
//#define  NTEMP              8
//#define  NHEAT              8
#endif // BASE

#ifdef H2RG
//#define  NBIAS              16
//#define  NCLKS              0
//#define  NTEMP              2
//#define  NHEAT              2
#endif // H2RG

/* Image descriptor structure definitions */
typedef struct  {
    u_short *pixbuf;            /* pixel buffer pointer */
    u_short *pix[8];            /* pixel buffer pointers */
    int bitpix;                 /* BITS PER PIXEL */
    int naxis;                  /* NUMBER OF AXES */
    int ncol;                   /* NAXIS1 */
    int nrow;                   /* NAXIS2 */
    int nccd;                   /* NAXIS3 */
    int nch;                    /* NAXIS3 */
    char seq_mode[40];          /* comment from seq modes file */
    char objtitle[80];          /* Object, dark, bias, flat, etc */
    char dateobs[20];           /* date (dd/mm/yy) of observation */
    char utime[20];             /* universal time @ start */
    char exptime[20];           /* actual integration time */
    } IMDESC;

/* Sequencer mode definition structure */
typedef struct {
    int mode;           /* Sequencer modes for readout */
    int nrow;           /* Number of rows */
    int ncol;           /* Number of columns */
    int nft;            /* Number of rows for frame transfer */
    int nflush;         /* Number of rows to flush */
//  int nch;            /* Number of output/readout channels */
    int nccd;           /* Number of detectors */
    int namps;          /* Number of output amplifiers */
    int npix;           /* Number of pixels */
    int vbin;           /* Vertical (parallel) binning */
    int hbin;           /* Horizontal (serial) binning */
    int vskip;          /* Prescan or offset */
    int hskip;          /* Prescan or offset */
    int vsub;           /* V subraster size */
    int hsub;           /* H subraster size */
    int nshuffle;       /* Number of shuffle-rows */
    int seqtime;        /* Sequencer exposure time (ms) */
    int gain;           /* Signal processing gain switch */
    int usec;           /* Signal processing sampling time */
    int opch;           /* Output channel selection */
    int sclks;          /* Serial clocks direction */
    int pid;            /* PID temperature control on/off */
    double tec[2];      /* Thermo-electric cooler current */
    double tset[8];     /* CCD temperature set-point(s) */
    double temp[8];     /* CCD temperatures */
    double heat[8];     /* CCD heater currents */
    double vclk[32];    /* Clock voltages */
    double vbias[16];   /* Bias voltages */
    } SEQ_MODE;

/* Parser command table entry structure */
typedef struct
    {
    char *nmem;             /* command mnemonic */
    void  (*fcn)();         /* pointer to associated function */
    } CMD_ENTRY;


#ifdef I386
#define BIGxENDIAN  0
#endif

#ifdef SPARC
#define BIGxENDIAN 1
#endif


/*============================================================================*/
/*     Function definitions for MOTXER application                            */
/*============================================================================*/

#ifndef MOTXER_C
extern void check_timer(void);
#endif

#include "cmddict.h"
#include "ccdutils.h"
#include "windex.h"
#include "plot.h"
#include "plotsub.h"
#include "plotdev.h"
#include "dspdvr.h"

/*============================================================================*/
