/*===========================================================================*/
/* Plot routines                                                             */
/*===========================================================================*/
#define PLOT_C

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <string.h>
#include "motxer.h"


/*============================================================================*/
#define  MAXHISTO           500
#define  MAX_REGSIZE        5000

#define  PF_MAXXMAX         MAX_REGSIZE
#define  PF_MINXRA          20
#define  MAXPLOTWIN         2

#define  LO_MINVAL          400
#define  LO_MAXVAL          5000
#define  HI_MINVAL          4000
#define  HI_MAXVAL          33000

#define  NPNTS              32


/* structure for plot storage and manipulation */
typedef struct plotdesc PLOTDESC;
struct plotdesc
    {
    PLOTDESC *plot_next;
    int      plot_size;     /* number of data points in plot_buff */
    int      plot_xlow;     /* x value of first point, 00 for spec */
    char     *plot_xlbl;
    char     *plot_ylbl;
    double   *plot_buff;
    };


/*========== Global variables ================================================*/
static int pf_ymin = -1;
static int pf_ymax = -1;
static int pf_xmin = -1;
static int pf_xmax = -1;

PLOTDESC  cx_plotdesc;

/*========== Function prototypes =============================================*/
static PLOTDESC* pf_getplotdesc(int bufsize, int newbox);
static int free_plot_bufs();
static void pf_plotexec(PLOTDESC *pd, int newbox);
static int vmsdf(double *src, int snx, double *mean, double *sigma, 
    double *min, double *max);
static void bitsplot(int minval, int maxval, int maxcount, double *valttl, 
    double *bitval, int npts);
static void bits_exec(int minval, int maxval);

/*----------------------------------------------------------------------------*/

/* alloc space for a plot data buffer,  free any unneeded descriptors */    
static PLOTDESC*
pf_getplotdesc(int bufsize, int newbox)
{
    extern PLOTDESC cx_plotdesc;
    PLOTDESC *lpd, *npd, *pd=&cx_plotdesc;
    double *buf;

    if(newbox) {
        npd  = pd->plot_next;   
        while(npd != NULL) {
            lpd = npd->plot_next;
            free(npd->plot_buff);
            free(npd->plot_xlbl);
            free(npd->plot_ylbl);
            free(npd);
            npd = lpd;
        }

        if(bufsize > pd->plot_size || pd->plot_buff == NULL) {
            free(pd->plot_buff);
            pd->plot_buff=(double*)malloc(bufsize*sizeof(double));
        }
        if(pd->plot_buff == NULL) {
            printf("GETPLOTDESC: memory allocation error\n");
            return(NULL);
        }
    }
    else {
        pd   = (PLOTDESC*)malloc(sizeof(PLOTDESC));
        buf  = (double*)malloc(bufsize*sizeof(double));
        if(buf==NULL || pd==NULL) {
            printf("GETPLOTDESC: memory allocation error\n");
            return(NULL);
        }
        pd->plot_buff =  buf;

        lpd = &cx_plotdesc;
        while(lpd->plot_next != NULL) lpd = lpd->plot_next;  
        lpd->plot_next = pd;
    }
    pd->plot_next = NULL;
    pd->plot_size = bufsize;
    pd->plot_xlow = 0;
    pd->plot_xlbl = NULL;
    pd->plot_ylbl = NULL;

    return(pd);
}


/* get rid of unneeded plot descriptors and buffers */  
static int 
free_plot_bufs(void)
{
    int numpd;
    PLOTDESC *pd, *npd;
    extern PLOTDESC cx_plotdesc;

    npd  = &cx_plotdesc;    
    pd  = npd->plot_next;        /* default PLOTDESC is permanent */

    numpd = 0;
    while(pd != NULL) {
        if(pd->plot_xlbl != NULL) free(pd->plot_xlbl);
        if(pd->plot_ylbl != NULL) free(pd->plot_ylbl);
        if(pd->plot_buff != NULL) free(pd->plot_buff);
    
        npd = pd->plot_next;
        free(pd);
        numpd++;
        pd = npd;
    }
}


/* set new Xlo->Xhi part of WCS for a row or column plot */
void
pf_setxrange(int xlo, int xhi) 
{
    extern PLOTDESC cx_plotdesc;
    extern int pf_xmin, pf_xmax;
    int t, pd_xlo, pd_xhi;
    PLOTDESC *pd = &cx_plotdesc;

    pd_xlo = pd->plot_xlow;
    pd_xhi = pd_xlo + pd->plot_size;

    if(xlo>=0 && xhi>=0) {      /* if not setting to auto, check range */
        if(xhi<xlo)
            { t = xhi; xhi = xlo; xlo = t; }
        if(xlo > (PF_MAXXMAX - PF_MINXRA)) {
            printf("pf_setxrange: plot arg or regon is out of range\n");
            return;
        }

        if (xlo < 0) xlo = 0;
        if (xhi<0) xhi=0; 
        if (xhi>PF_MAXXMAX)xhi=PF_MAXXMAX;
        if(xhi<(xlo+PF_MINXRA)) xhi=(xlo+PF_MINXRA);
        if (xlo<pd_xlo) xlo=pd_xlo;
        if (xhi>pd_xhi) xhi=pd_xhi;
    }
    pf_xmin = xlo;
    pf_xmax = xhi;
/*  printf("new xlo = %d, new xhi = %d\n", xlo, xhi); */

}


/* set Y part of WCS for a row or column plot */
void
pf_setyrange(int min, int max)
{
    extern int pf_ymin, pf_ymax;

    pf_ymin = min;
    pf_ymax = max;
}


/*  plot plot_desc contents */
static void
pf_plotexec(PLOTDESC *pd, int newbox)
{
    extern int pf_xmin, pf_xmax;
    extern int pf_ymin, pf_ymax;
    static int color;
    int regsiz, pix1, npix, xlow;
    int i;
    double *xbuf, *ybuf, min, max, mean, sdev;
    static char label[200];

    regsiz  = pd->plot_size;
    xlow    = pd->plot_xlow;

    /* get the current x-range values */
    if(pf_xmin<0 || pf_xmax<0){                 /* if auto-range on x */
        pix1 = xlow;                        /* use values from plot desc */
        npix = regsiz;
    }
    else {                              /* else use the set values */
        pix1 = pf_xmin;
        npix = (pf_xmax - pf_xmin) + 1;
    }

    /* check for range errors */
    if(pix1 > regsiz || pix1 < xlow  || (pix1+npix) < xlow) {
        pix1 = xlow;
        npix = regsiz;
        pf_setxrange(pix1, (pix1+npix));
    }
    if((pix1+npix) > (regsiz+xlow)) npix = regsiz + xlow - pix1;

    xbuf   = (double*)malloc(npix*sizeof(double));
    if(xbuf == NULL) return;
    ybuf = pd->plot_buff + (pix1 - xlow);


    /* get statistics on what is to be displayed */
    vmsdf(ybuf, npix, &mean, &sdev, &min, &max);

    if(pd->plot_xlbl == NULL){
        sprintf(label, "min %d, max %d, mean %.1f, sdev %.2f",
            (int)min, (int)max, mean, sdev);
    }
    else {
        strcpy(label, pd->plot_xlbl);
    }

    if(newbox)  {
        if(pf_ymin<0 || pf_ymax<0) {        /* if auto-range on y */
            max += max/50;
            min -= max/50;
            if (min<0)min=0; 
        }
        else {                          /* use the set values */
            min = pf_ymin;
            max = pf_ymax;
        }
    }

    /* set the x values */
    for(i=0; i<npix; i++) xbuf[i] = (double)(i+pix1);
 
    if(newbox) 
        color = 2;
    else         
        color++;    
 
    if(newbox) {
        psreset();
        psplotmode();
        pserase();
        pssetmarg(3.0, 0.2, 1.7, 0.2);
        pssetarea(0.0, 0.0, 1.0, 1.0);
        pssetwcs((double)(pix1), (double)min,
                 (double)(pix1+npix), (double)max);
        psbox("bctsn", "bctsn");
    }
    pssetcolor(color);


    if(npix>MAXHISTO)
        psconnect(xbuf, ybuf, npix);
    else
        pshisto(xbuf, ybuf, npix);

    if(newbox) psboxlabels(label, "");
    psgflush();

    free (xbuf);
}

static int
vmsdf(double *src, int snx, double *mean, double *sigma, double *min,
    double *max) 
{
    int i,n;
    double val, pix;
    double sum=0.0, sum2=0.0;

    n = snx/3;
    pix = (double)n;
    *min = src[0];
    *max = src[0];

    for(i=0; i<snx; i++) {
        val   = src[i];
        if(val>*max) *max = val;
        if(val<*min) *min = val;

        // do mean, sigma on middle 1/3 of line only
        if (i>n && i<=2*n){
            sum  += val;
            sum2 += val*val;
        }
    }

    *mean = sum / pix;
    *sigma = sqrt((sum2-(sum*sum)/pix)/(pix-1.0));

    return((int)pix);
}

/*  backup copy
static int
vmsdf(double *src, int snx, double *mean, double *sigma, double *min,
    double *max) 
{
    int i;
    double val, pix;
    double sum=0.0, sum2=0.0;

    pix = (double)snx;
    *min = src[0];
    *max = src[0];

    for(i=0; i<snx; i++) {
        val   = src[i];
        if(val>*max) *max = val;
        if(val<*min) *min = val;

        sum  += val;
        sum2 += val*val;
    }

    *mean = sum / pix;
    *sigma = sqrt((sum2-(sum*sum)/pix)/(pix-1.0));

    return((int)pix);
}
*/

/* plot a row */
void
pf_plotrow(IMDESC *im, int row, int newbox)
{
    u_short *src;               /* pointer to source array */
    double *dst;                /* pointer to destination array */
    int i, j;
    PLOTDESC *pd;

    if (row>im->nrow || im->nrow < 1 || im->ncol < 1){
        printf("Plot arg is out of range \07\n");
        return;
    }

    if((pd = pf_getplotdesc(im->ncol, newbox)) == NULL) return;

    src = im->pixbuf;
    dst = pd->plot_buff;
    for (j=0; j<im->ncol; j++)dst[j] = (double)src[row*im->ncol+j];

    pf_plotexec(pd, newbox);
}


/*  plot a column */
void
pf_plotcol(IMDESC *im, int col, int newbox)
{
    u_short *src;               /* pointer to source array */
    double *dst;                /* pointer to destination array */
    int i, j;
    PLOTDESC *pd;

    if (col>im->ncol || im->nrow < 1 || im->ncol < 1){
        printf("Plot arg is out of range \07\n");
        return;
    }

    if((pd = pf_getplotdesc(im->nrow, newbox)) == NULL) return;

    src = im->pixbuf;
    dst = pd->plot_buff;
    for (j=0; j<im->nrow; j++)dst[j] = (double)src[col+j*im->ncol];

    pf_plotexec(pd, newbox); 
}

/*  replot active plotbufs with current plotwindow */
void
pf_replot(void)
{
    extern PLOTDESC cx_plotdesc;
    PLOTDESC *pd;
    int previous, newbox;

    pd = &cx_plotdesc;
    newbox = TRUE;                  /* first plotted sets y-range. */

    pf_plotexec(pd , newbox);
    newbox = FALSE;

    while((pd = pd->plot_next) != NULL) pf_plotexec(pd , newbox);
}


/* plot npix from row starting at col  */
static void
bitsplot(int minval, int maxval, int maxcount, double *valttl, 
    double *bitval, int npts)
{
    int rsiz;
    double *sp, *dp; 
    PLOTDESC *dst, *pf_getplotdesc();
    static char *xlbl = "ADC value distribution";
    rsiz = npts;

    if((dst = pf_getplotdesc(rsiz, TRUE)) == NULL) return;

    sp = valttl;
    dp = dst->plot_buff;
    do  {
        *dp++ = *sp++;
        } while(--rsiz>0);

    dst->plot_size = npts;
    dst->plot_xlow = minval;
    dst->plot_xlbl = xlbl;

    pf_plotexec(dst, TRUE);
}

static void
bits_exec(int minval, int maxval)
{
    extern IMDESC *im;
    ushort *zbuf;
    int i, npix;
    ushort val;
    double *bitval;
    double *valttl;
    double maxcount;
    int npts;

    npts = maxval - minval + 1;

    bitval = (double*)malloc(npts*sizeof(double));
    valttl = (double*)malloc(npts*sizeof(double));

    zbuf = im->pixbuf;
    npix = (im->ncol) * (im->nrow);
    for(i=0; i<npts; i++) {         /* zero the totals buffer */
        valttl[i] = 0.0;
    }
    for(i=0; i<npix; i++) {         /* get the distribution totals */
        val = *zbuf++;
        if(val>minval && val<(ushort)maxval) valttl[ val-minval ] += 1.0;
    }
    for(i=0; i<npts; i++) {         /* set the ordinate values */
        bitval[i] = (double)(i+minval);
    }

    for(i=0, maxcount=0.0, val=0; i<npts; i++) {
        if(valttl[i] > maxcount) {
            maxcount = valttl[i];
            val = i + minval;
        }
    }
    printf("maxcount = %d @ value = %d\n", (int)maxcount, val);

    bitsplot(minval, maxval, (int)maxcount, valttl, bitval, npts);

    free (bitval);
    free (valttl);
}


void
lobits_plot(void)
{
    int minval, maxval;

    minval = LO_MINVAL;
    maxval = LO_MAXVAL;
    bits_exec(minval, maxval);
}


void
hibits_plot(void)
{
    int minval, maxval;

    minval = HI_MINVAL;
    maxval = HI_MAXVAL;
    bits_exec(minval, maxval);
}

void
bitbits_plot(void)
{
    int min, max, r;

    min = (int)fp_arg();
    max = (int)fp_arg();

    if(min < 0 || max < 0) {
        printf("Usage: bitbits lo-row hi-row\n");
        return;
    }

    if(max<min) { r = max; max = min; min = r; }
    bitbits(min, max-min);
}


void
bitbits(int firstrow, int nrow)
{
    extern IMDESC *im;
    u_int bittotels[NPNTS];         /* 16 bits, 16 zeros, 16 ones */
    int i, j, npix;
    ushort *zbuf, val, bt;

    if((firstrow+nrow) > im->ncol)  /* !!!!!!! */ return;

    zbuf = im->pixbuf;
    zbuf += (im->ncol)*firstrow;
    npix  = (im->ncol)*nrow;

    /* zero the totals buffer */
    for(i=0; i<NPNTS; i++) bittotels[i] = 0;

    for(i=0; i<npix; i++) {             /* get the distribution totals */
        val = *zbuf++;
        bt = 01; 
        for(j=0; j<NPNTS; j+=2) {
            if(!(val&bt)) 
                bittotels[j]++;         /* its a 0 */
            else 
                   bittotels[j+1]++;        /* its a 1 */
        bt <<= 1;
        }   
    }

    for(i=0; i<(NPNTS/2); i++) {        /* print the distribution totals */
        printf("Bit %2d: zeros= %6u, ones= %6u, ratio= %5.3f\n",
             i, bittotels[i*2], bittotels[i*2+1],
             (double)((double)bittotels[i*2]/(double)bittotels[i*2+1]));
    }
}
