/*===========================================================================*/
/* Plot support routines                                                     */
/*===========================================================================*/
#define PLOTSUB_C

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <sys/types.h>
#include <string.h>
#include "motxer.h"


/*========== definitions ====================================================*/
#define NOPTIONS        5

/* The offsets in the following #defines must correspond to
 * the appropriate option symbol in optsymbols[].
 */
#define  X_STDSIDE      (xflags[0])
#define  X_ALTSIDE      (xflags[1])
#define  X_BIGTIC       (xflags[2])
#define  X_SMLTIC       (xflags[3])
#define  X_STDLBL       (xflags[4])

#define  Y_STDSIDE      (yflags[0])
#define  Y_ALTSIDE      (yflags[1])
#define  Y_BIGTIC       (yflags[2])
#define  Y_SMLTIC       (yflags[3])
#define  Y_STDLBL       (yflags[4])


/* default plot area, excluding space for labels, titles etc. */
#define  XLOSC      3200L           /* Screen origin of boxes is (X1,Y1) */
#define  YLOSC      3200L
#define  XHISC      29567L          /* Screen top right of boxes is (X2,Y2) */
#define  YHISC      29567L
#define  SC_MAX     32767L          /* plot window size in screen coord*/

/* output device dependent functions */
typedef struct {
   char *dev_name;                  /* name of device */
   int (*dev_open)();               /* initialization  */
   int (*dev_plotmode)();           /* enable graphics */
   int (*dev_line)();               /* draw a line from (x1,y1) to (x2,y2) */
   int (*dev_dot)();                /* draw a single pixel dot */
   int (*dev_ltype)();              /* hardware line type */
   int (*dev_lwidth)();             /* hardware line weight */
   int (*dev_erase)();              /* erase graphics screen  */
   int (*dev_rdcursor)();           /* cursor display */
   int (*dev_color)();              /* set colour of line */
   int (*dev_gflush)();             /* flush graphics */
   int (*dev_textmode)();           /* switch from graph to alpha mode */
   int (*dev_close)();              /* close device */
} DEVICE;


#define  DEV_NODEVICE       0
#define  DEV_X11            1

/*  defines for graphics functions  */
#define  DEV_OPEN           (*devices[ps_device].dev_open)
#define  DEV_PLOTMODE       (*devices[ps_device].dev_plotmode)
#define  DEV_LINE           (*devices[ps_device].dev_line)
#define  DEV_DOT            (*devices[ps_device].dev_dot)
#define  DEV_LTYPE          (*devices[ps_device].dev_ltype)
#define  DEV_LWIDTH         (*devices[ps_device].dev_lwidth)
#define  DEV_ERASE          (*devices[ps_device].dev_erase)
#define  DEV_RDCURSOR       (*devices[ps_device].dev_rdcursor)
#define  DEV_COLOR          (*devices[ps_device].dev_color)
#define  DEV_GFLUSH         (*devices[ps_device].dev_gflush)
#define  DEV_TEXTMODE       (*devices[ps_device].dev_textmode)
#define  DEV_CLOSE          (*devices[ps_device].dev_close)


#define  AXISLBLGAP     2.2

#define  SWAP_PTS(X0, Y0, CODES0, X1, Y1, CODES1)   \
    tmpl = *X0; *X0 = *X1; *X1 = tmpl;              \
    tmpl = *Y0; *Y0 = *Y1; *Y1 = tmpl;              \
    tmpi = CODES0; CODES0 = CODES1; CODES1 = tmpi;


#define  OUT_LEFT       8
#define  OUT_RIGHT      4
#define  OUT_BOT        2
#define  OUT_TOP        1

#define  PSOPEN         1

/* defaults for plotting aesthetically pleasing text */
#define  GLFMAG         25          /* default font magnification, = 1.0 */
#define  GLFTALL        32          /* unenlarged character height */
#define  GLFWIDE        20          /* unenlarged character width */

#define  GLFBASE        12
#define  SUBOFF         20

#define  CHTALL         (double)(GLFTALL*GLFMAG)

struct glfspec
    {
    u_char radj;            /* right adjustment of glyph */
    u_char ladj;            /* left adjustment of glyph */
    u_char npoints;         /* # points defining the glyph  */
    int offset;             /* offset into stroke table */
    };

/*========== function prototypes ============================================*/
static void drwxticks(void);
static void drwyticks(void);
static void chkoptions(char *str, char *opflgs);
static double psround(double v, int *nsub);
static void settextsizes(void);
static void pssetscale(void);
static void pstxtscale(double arg);
static void pstxtaspect(double arg);
static void pssetangle(int arg);
static void psstrsc(int xpos, int ypos, char *str);
static int pstiklabel(double val, char *clbl);
static int no_open(void);
static int no_plotmode(void);
static int no_line(void);
static int no_dot(void);
static int no_ltype(void);
static int no_lwidth(void);
static int no_erase(void);
static int no_cursor(void);
static int no_color(void);
static int no_gflush(void);
static int no_textmode(void);
static int no_close(void);
static int gxline(int x0, int y0, int x1, int y1, char limitflag);
static void gxstring(int xpos, int ypos, char *str, int drwflg,
    int *str_dx, int *str_dy);
static int gxstrsize(char *str, char axis);

/*========== local variables ================================================*/
static char optsymbols[NOPTIONS] = { 'b', 'c', 't', 's', 'n' };

static  char xflags[NOPTIONS];
static  char yflags[NOPTIONS];

static double lintbl[10] =
    { 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9 };

int ps_botmarg      = 3200L;        /* margins from current window edges */
int ps_topmarg      = 3200L;        /* to the plot area. Set In NDC. */
int ps_leftmarg     = 3200L;        /* space for labels, titles, etc. */
int ps_rightmarg    = 3200L;

double ps_xlowc     = 0.0;          /* User coords of plot area */
double ps_xhiwc     = 1.0;
double ps_ylowc     = 0.0;
double ps_yhiwc     = 1.0;

double ps_xscale = (double)(XHISC - XLOSC);
double ps_yscale = (double)(YHISC - YLOSC);
double ps_xoffset = (double)XLOSC;
double ps_yoffset = (double)YHISC;

int ps_xlosc    = XLOSC;            /* Device coords of plot area */
int ps_xhisc    = XHISC;
int ps_ylosc    = YLOSC;
int ps_yhisc    = YHISC;
int ps_isopen   = 0;                /* psopen must be called first */
int ps_device   = 1;                /* Current device */
int ps_ltype    = 0;                /* Line type: dotted; etc. */
int ps_lwide    = 1;                /* Line width: 1;2...etc */
int ps_color    = 1;                /* */
int ps_clip     = 1;                /* clipping flag */
int ldef        = 32;               /* scale spacing for ps_lwide > 1 lines */
int ps_angle    = 0;                /* rotation of chars and points */
int ps_cosa     = 1;                /* cos(ps_angle) */
int ps_sina     = 0;                /* sin(ps_angle) */
int ps_xtxtscal = GLFMAG;           /* scale factors for text, symbols, ticks */
int ps_ytxtscal = GLFMAG;
int ps_xchgap = (GLFTALL*GLFMAG);   /* spacing for labels from axis, etc */
int ps_ychgap = (GLFTALL*GLFMAG);   /* adjusted for aspect ratio */

static char clbl[22];

DEVICE devices[] =
  {
    {
    "nodevice",                     /* nodevice - must be first */
    no_open, no_plotmode, no_line, no_dot, no_ltype,
    no_lwidth, no_erase,  no_cursor,
    no_color, no_gflush, no_textmode, no_close
    },
    {
    "X",                            /* plot with X */
    px_open, px_plotmode, px_line, no_dot, no_ltype,
    px_lwidth, px_erase, px_rdcursor,
    px_color, px_gflush, px_textmode, px_close,
    }
  };

int ndev = (sizeof(devices)/sizeof(DEVICE));        /* number of devices */

static double nice[3] = { 2.0, 5.0, 10.0 };

static double ps_textscale  = 1.0;
static double ps_textaspect = 1.0;                  /* ysiz / xsiz */

#define     NCHAR 127                               /* chars per font */
struct glfspec simplex_index[NCHAR] =
    {
    {  7,  7, 17,  1410}, {  6,  6,  5,  1444}, {  7,  7,  4,  1454},
    {  6,  6,  5,  1462}, {  8,  8, 11,  1472}, {  7,  7,  4,  1494},
    {  5,  5,  4,  1502}, {  5,  5,  6,  1510}, {  4,  4, 27,  1522},
    {  4,  4, 19,  1576}, {  5,  5, 12,  1614}, {  3,  6, 12,  1638},
    {  5,  5, 12,  1662}, {  6,  3, 12,  1686}, {  6,  6, 16,  1710},
    {  6,  6, 13,  1742}, {  7,  7, 30,  1768}, {  7,  7,  8,  1828},
    {  1,  1,  5,  1844}, { 10, 10, 25,  1868}, {  6,  6, 36,  1918},
    {  6,  6, 24,  1990}, {  6,  6, 38,  2038}, { 10, 10, 22,  2114},
    {  7,  6, 27,  2158}, { 10, 10, 21,  2212}, {  7,  6, 18,  2254},
    { 10, 10,  8,  3275}, {  9, 11,  7,  1854}, { 12, 12, 14,  2290},
    { 11, 10, 15,  2318}, {  8,  8,  0,  2348}, {  5,  5, 12,  2417},
    {  8,  8,  8,  2499}, { 11, 10,  8,  3259}, { 10, 10, 38,  3183},
    { 12, 12, 29,  2929}, { 13, 12, 46,  2987}, {  5,  5,  7,  2901},
    {  7,  7, 18,  2557}, {  7,  7, 18,  2593}, {  8,  8,  6,  2541},
    { 13, 13,  4,  2817}, {  5,  5,  7,  2359}, { 13, 13,  2,  2813},
    {  5,  5,  5,  2349}, { 11, 11,  2,  2553}, { 10, 10, 17,  1104},
    { 10, 10,  4,  1138}, { 10, 10, 14,  1146}, { 10, 10, 15,  1174},
    { 10, 10,  5,  1204}, { 10, 10, 17,  1214}, { 10, 10, 23,  1248},
    { 10, 10,  4,  1294}, { 10, 10, 29,  1302}, { 10, 10, 23,  1360},
    {  5,  5, 10,  2373}, {  5,  5, 12,  2393}, { 12, 12,  3,  2833},
    { 13, 13,  4,  2825}, { 12, 12,  3,  2839}, {  9,  9, 29,  2441},
    { 14, 13, 52,  3079}, {  9,  9,  6,     0}, { 10, 11, 21,    12},
    { 11, 10, 18,    54}, { 10, 11, 14,    90}, {  9, 10,  8,   118},
    {  8, 10,  6,   134}, { 11, 10, 21,   146}, { 11, 11,  6,   188},
    {  4,  4,  2,   200}, {  8,  8, 10,   204}, { 10, 11,  6,   224},
    {  7, 10,  4,   236}, { 12, 12,  8,   244}, { 11, 11,  6,   260},
    { 11, 11, 21,   272}, { 10, 11, 12,   314}, { 11, 11, 23,   338},
    { 10, 11, 14,   384}, { 10, 10, 20,   412}, {  8,  8,  4,   452},
    { 11, 11, 10,   460}, {  9,  9,  4,   480}, { 12, 12,  8,   488},
    { 10, 10,  4,   504}, {  9,  9,  5,   512}, { 10, 10,  6,   522},
    {  7,  7,  8,  2629}, {  7,  7,  2,  1406}, {  7,  7,  8,  2645},
    { 11, 11,  6,  2889}, {  8,  7,  2,  3291}, {  5,  5,  7,  2915},
    { 10,  9, 16,   534}, {  9, 10, 16,   566}, {  9,  9, 14,   598},
    { 10,  9, 16,   626}, {  9,  9, 17,   658}, {  7,  5,  7,   692},
    { 10,  9, 21,   706}, { 10,  9,  9,   748}, {  4,  4,  7,   766},
    {  5,  5, 10,   780}, {  8,  9,  6,   800}, {  4,  4,  2,   812},
    { 15, 15, 16,   816}, { 10,  9,  9,   848}, { 10,  9, 17,   866},
    {  9, 10, 16,   900}, { 10,  9, 16,   932}, {  6,  7,  7,   964},
    {  9,  8, 17,   978}, {  7,  5,  7,  1012}, { 10,  9,  9,  1026},
    {  8,  8,  4,  1044}, { 11, 11,  8,  1052}, {  9,  8,  4,  1068},
    {  8,  8,  8,  1076}, {  9,  8,  6,  1092}, {  7,  7, 37,  2661},
    {  4,  4,  2,  2809}, {  7,  7, 37,  2735}, { 12, 12, 22,  2845},
    {  7,  7, 13,  2515},
    };

char simplex_table[] =
    {
    210, 70, 74, 91,210, 70, 90, 91,205, 84, 87, 84,203, 70, 75, 91,203, 70, 84,
     70, 87, 71, 88, 72, 89, 74, 89, 76, 88, 78, 87, 79, 84, 80,203, 80, 84, 80,
     87, 81, 88, 82, 89, 84, 89, 87, 88, 89, 87, 90, 84, 91, 75, 91,218, 75, 89,
     73, 87, 71, 85, 70, 81, 70, 79, 71, 77, 73, 76, 75, 75, 78, 75, 83, 76, 86,
     77, 88, 79, 90, 81, 91, 85, 91, 87, 90, 89, 88, 90, 86,203, 70, 75, 91,203,
     70, 82, 70, 85, 71, 87, 73, 88, 75, 89, 78, 89, 83, 88, 86, 87, 88, 85, 90,
     82, 91, 75, 91,204, 70, 76, 91,204, 70, 89, 70,204, 80, 84, 80,204, 91, 89,
     91,204, 70, 76, 91,204, 70, 89, 70,204, 80, 84, 80,218, 75, 89, 73, 87, 71,
     85, 70, 81, 70, 79, 71, 77, 73, 76, 75, 75, 78, 75, 83, 76, 86, 77, 88, 79,
     90, 81, 91, 85, 91, 87, 90, 89, 88, 90, 86, 90, 83,213, 83, 90, 83,203, 70,
     75, 91,217, 70, 89, 91,203, 80, 89, 80,210, 70, 82, 91,214, 70, 86, 86, 85,
     89, 84, 90, 82, 91, 80, 91, 78, 90, 77, 89, 76, 86, 76, 84,203, 70, 75, 91,
    217, 70, 75, 84,208, 79, 89, 91,204, 70, 76, 91,204, 91, 88, 91,202, 70, 74,
     91,202, 70, 82, 91,218, 70, 82, 91,218, 70, 90, 91,203, 70, 75, 91,203, 70,
     89, 91,217, 70, 89, 91,208, 70, 78, 71, 76, 73, 75, 75, 74, 78, 74, 83, 75,
     86, 76, 88, 78, 90, 80, 91, 84, 91, 86, 90, 88, 88, 89, 86, 90, 83, 90, 78,
     89, 75, 88, 73, 86, 71, 84, 70, 80, 70,203, 70, 75, 91,203, 70, 84, 70, 87,
     71, 88, 72, 89, 74, 89, 77, 88, 79, 87, 80, 84, 81, 75, 81,208, 70, 78, 71,
     76, 73, 75, 75, 74, 78, 74, 83, 75, 86, 76, 88, 78, 90, 80, 91, 84, 91, 86,
     90, 88, 88, 89, 86, 90, 83, 90, 78, 89, 75, 88, 73, 86, 71, 84, 70, 80, 70,
    211, 87, 89, 93,203, 70, 75, 91,203, 70, 84, 70, 87, 71, 88, 72, 89, 74, 89,
     76, 88, 78, 87, 79, 84, 80, 75, 80,210, 80, 89, 91,217, 73, 87, 71, 84, 70,
     80, 70, 77, 71, 75, 73, 75, 75, 76, 77, 77, 78, 79, 79, 85, 81, 87, 82, 88,
     83, 89, 85, 89, 88, 87, 90, 84, 91, 80, 91, 77, 90, 75, 88,210, 70, 82, 91,
    203, 70, 89, 70,203, 70, 75, 85, 76, 88, 78, 90, 81, 91, 83, 91, 86, 90, 88,
     88, 89, 85, 89, 70,202, 70, 82, 91,218, 70, 82, 91,200, 70, 77, 91,210, 70,
     77, 91,210, 70, 87, 91,220, 70, 87, 91,203, 70, 89, 91,217, 70, 75, 91,202,
     70, 82, 80, 82, 91,218, 70, 82, 80,217, 70, 75, 91,203, 70, 89, 70,203, 91,
     89, 91,216, 77, 88, 91,216, 80, 86, 78, 84, 77, 81, 77, 79, 78, 77, 80, 76,
     83, 76, 85, 77, 88, 79, 90, 81, 91, 84, 91, 86, 90, 88, 88,204, 70, 76, 91,
    204, 80, 78, 78, 80, 77, 83, 77, 85, 78, 87, 80, 88, 83, 88, 85, 87, 88, 85,
     90, 83, 91, 80, 91, 78, 90, 76, 88,216, 80, 86, 78, 84, 77, 81, 77, 79, 78,
     77, 80, 76, 83, 76, 85, 77, 88, 79, 90, 81, 91, 84, 91, 86, 90, 88, 88,216,
     70, 88, 91,216, 80, 86, 78, 84, 77, 81, 77, 79, 78, 77, 80, 76, 83, 76, 85,
     77, 88, 79, 90, 81, 91, 84, 91, 86, 90, 88, 88,204, 83, 88, 83, 88, 81, 87,
     79, 86, 78, 84, 77, 81, 77, 79, 78, 77, 80, 76, 83, 76, 85, 77, 88, 79, 90,
     81, 91, 84, 91, 86, 90, 88, 88,215, 70, 85, 70, 83, 71, 82, 74, 82, 91,207,
     77, 86, 77,216, 77, 88, 93, 87, 96, 86, 97, 84, 98, 81, 98, 79, 97,216, 80,
     86, 78, 84, 77, 81, 77, 79, 78, 77, 80, 76, 83, 76, 85, 77, 88, 79, 90, 81,
     91, 84, 91, 86, 90, 88, 88,205, 70, 77, 91,205, 81, 80, 78, 82, 77, 85, 77,
     87, 78, 88, 81, 88, 91,209, 70, 82, 71, 83, 70, 82, 69, 81, 70,210, 77, 82,
     91,210, 70, 83, 71, 84, 70, 83, 69, 82, 70,211, 77, 83, 94, 82, 97, 80, 98,
     78, 98,205, 70, 77, 91,215, 77, 77, 87,209, 83, 88, 91,210, 70, 82, 91,199,
     77, 71, 91,199, 81, 74, 78, 76, 77, 79, 77, 81, 78, 82, 81, 82, 91,210, 81,
     85, 78, 87, 77, 90, 77, 92, 78, 93, 81, 93, 91,205, 77, 77, 91,205, 81, 80,
     78, 82, 77, 85, 77, 87, 78, 88, 81, 88, 91,209, 77, 79, 78, 77, 80, 76, 83,
     76, 85, 77, 88, 79, 90, 81, 91, 84, 91, 86, 90, 88, 88, 89, 85, 89, 83, 88,
     80, 86, 78, 84, 77, 81, 77,204, 77, 76, 98,204, 80, 78, 78, 80, 77, 83, 77,
     85, 78, 87, 80, 88, 83, 88, 85, 87, 88, 85, 90, 83, 91, 80, 91, 78, 90, 76,
     88,216, 77, 88, 98,216, 80, 86, 78, 84, 77, 81, 77, 79, 78, 77, 80, 76, 83,
     76, 85, 77, 88, 79, 90, 81, 91, 84, 91, 86, 90, 88, 88,207, 77, 79, 91,207,
     83, 80, 80, 82, 78, 84, 77, 87, 77,216, 80, 87, 78, 84, 77, 81, 77, 78, 78,
     77, 80, 78, 82, 80, 83, 85, 84, 87, 85, 88, 87, 88, 88, 87, 90, 84, 91, 81,
     91, 78, 90, 77, 88,210, 70, 82, 87, 83, 90, 85, 91, 87, 91,207, 77, 86, 77,
    205, 77, 77, 87, 78, 90, 80, 91, 83, 91, 85, 90, 88, 87,216, 77, 88, 91,204,
     77, 82, 91,216, 77, 82, 91,202, 77, 78, 91,210, 77, 78, 91,210, 77, 86, 91,
    218, 77, 86, 91,205, 77, 88, 91,216, 77, 77, 91,204, 77, 82, 91,216, 77, 82,
     91, 80, 95, 78, 97, 76, 98, 75, 98,216, 77, 77, 91,205, 77, 88, 77,205, 91,
     88, 91,209, 70, 78, 71, 76, 74, 75, 79, 75, 82, 76, 87, 78, 90, 81, 91, 83,
     91, 86, 90, 88, 87, 89, 82, 89, 79, 88, 74, 86, 71, 83, 70, 81, 70,206, 74,
     80, 73, 83, 70, 83, 91,204, 75, 76, 74, 77, 72, 78, 71, 80, 70, 84, 70, 86,
     71, 87, 72, 88, 74, 88, 76, 87, 78, 85, 81, 75, 91, 89, 91,205, 70, 88, 70,
     82, 78, 85, 78, 87, 79, 88, 80, 89, 83, 89, 85, 88, 88, 86, 90, 83, 91, 80,
     91, 77, 90, 76, 89, 75, 87,213, 70, 75, 84, 90, 84,213, 70, 85, 91,215, 70,
     77, 70, 76, 79, 77, 78, 80, 77, 83, 77, 86, 78, 88, 80, 89, 83, 89, 85, 88,
     88, 86, 90, 83, 91, 80, 91, 77, 90, 76, 89, 75, 87,216, 73, 87, 71, 84, 70,
     82, 70, 79, 71, 77, 74, 76, 79, 76, 84, 77, 88, 79, 90, 82, 91, 83, 91, 86,
     90, 88, 88, 89, 85, 89, 84, 88, 81, 86, 79, 83, 78, 82, 78, 79, 79, 77, 81,
     76, 84,217, 70, 79, 91,203, 70, 89, 70,208, 70, 77, 71, 76, 73, 76, 75, 77,
     77, 79, 78, 83, 79, 86, 80, 88, 82, 89, 84, 89, 87, 88, 89, 87, 90, 84, 91,
     80, 91, 77, 90, 76, 89, 75, 87, 75, 84, 76, 82, 78, 80, 81, 79, 85, 78, 87,
     77, 88, 75, 88, 73, 87, 71, 84, 70, 80, 70,216, 77, 87, 80, 85, 82, 82, 83,
     81, 83, 78, 82, 76, 80, 75, 77, 75, 76, 76, 73, 78, 71, 81, 70, 82, 70, 85,
     71, 87, 73, 88, 77, 88, 82, 87, 87, 85, 90, 82, 91, 80, 91, 77, 90, 76, 88,
    203, 70, 89, 94,209, 75, 78, 76, 76, 78, 75, 81, 75, 83, 76, 86, 78, 88, 81,
     89, 83, 89, 86, 88, 88, 86, 89, 83, 89, 81, 88, 78, 86, 76, 83, 75, 81, 75,
    204, 76, 76, 88, 88, 88, 88, 76, 76, 76,210, 74, 75, 86, 89, 86, 82, 74,210,
     72, 76, 82, 82, 92, 88, 82, 82, 72,210, 73, 80, 79, 74, 79, 79, 83, 77, 89,
     82, 85, 87, 89, 85, 83, 90, 79, 84, 79, 82, 73,210, 75, 82, 89,203, 82, 89,
     82,205, 77, 87, 87,215, 77, 77, 87,210, 76, 82, 88,205, 79, 87, 85,215, 79,
     77, 85,209, 78, 79, 79, 78, 81, 78, 83, 79, 85, 81, 86, 83, 86, 85, 85, 86,
     83, 86, 81, 85, 79, 83, 78, 81, 78,207, 81, 79, 83,208, 80, 80, 84,209, 79,
     81, 85,210, 79, 82, 85,211, 79, 83, 85,212, 80, 84, 84,213, 81, 85, 83,206,
     78, 78, 86, 86, 86, 86, 78, 78, 78,207, 79, 79, 85,208, 79, 80, 85,209, 79,
     81, 85,210, 79, 82, 85,211, 79, 83, 85,212, 79, 84, 85,213, 79, 85, 85,210,
     76, 77, 85, 87, 85, 82, 76,210, 79, 79, 84,210, 79, 85, 84,210, 82, 81, 84,
    210, 82, 83, 84,204, 82, 85, 87, 85, 77, 76, 82,207, 82, 84, 85,207, 82, 84,
     79,210, 82, 84, 83,210, 82, 84, 81,210, 88, 87, 79, 77, 79, 82, 88,210, 85,
     85, 80,210, 85, 79, 80,210, 82, 83, 80,210, 82, 81, 80,216, 82, 79, 77, 79,
     87, 88, 82,213, 82, 80, 79,213, 82, 80, 85,210, 82, 80, 81,210, 82, 80, 83,
    210, 76, 78, 87, 88, 80, 76, 80, 86, 87, 82, 76,210, 82, 82, 76,210, 82, 76,
     80,210, 82, 78, 87,210, 82, 86, 87,210, 82, 88, 80,208, 76, 80, 80, 76, 80,
     76, 84, 80, 84, 80, 88, 84, 88, 84, 84, 88, 84, 88, 80, 84, 80, 84, 76, 80,
     76,217, 80, 88, 78, 86, 76, 83, 75, 81, 75, 78, 76, 76, 78, 75, 81, 75, 83,
     76, 86, 78, 88, 81, 89, 83, 89, 86, 88, 88, 86, 89, 84,217, 80, 87, 78, 85,
     77, 83, 77, 81, 78, 80, 79, 79, 81, 79, 83, 80, 85, 81, 86, 83, 87, 85, 87,
     87, 86, 89, 84,210, 74, 75, 86, 89, 86, 82, 74,210, 90, 89, 78, 75, 78, 82,
     90,210, 81, 81, 82, 82, 83, 83, 82, 82, 81,201, 79, 76, 79, 82, 87,203, 79,
     82, 88,219, 70, 82, 88,218, 82, 89, 84, 87, 85, 86, 85, 84, 84, 83, 83, 81,
     80, 80, 79, 78, 78, 77, 78, 75, 79, 74, 81, 74, 82, 75, 84, 77, 85, 78, 85,
     80, 84, 81, 83, 83, 80, 84, 79, 86, 78, 87, 78, 89, 79, 90, 81, 90, 82,212,
     76, 83, 76, 83, 77, 84, 77, 84, 76, 83, 75, 81, 75, 80, 76, 80, 78, 81, 80,
     84, 82, 85, 83,208, 78, 81, 79, 84, 81, 85, 83, 85, 85, 83, 87,209, 80, 79,
     82, 79, 84, 80, 86, 83, 88, 84, 89,207, 84, 80, 85, 83, 87, 84, 89, 84, 91,
     83, 92, 81, 92, 80, 91, 80, 90, 81, 90, 81, 91, 80, 91,210, 75, 81, 76, 82,
     77, 83, 76, 82, 75,210, 77, 82, 81,210, 81, 81, 83, 82, 86, 83, 83, 82, 81,
    210, 86, 82, 92,208, 79, 79, 78, 78, 79, 79, 80, 80, 79, 84, 79, 85, 78, 86,
     79, 85, 80, 84, 79,210, 77, 83, 76, 82, 75, 81, 76, 82, 77, 82, 81, 81, 82,
     83, 85, 82, 86,210, 81, 83, 82, 81, 85, 82, 86, 82, 90, 81, 91, 82, 92, 83,
     91, 82, 90,208, 79, 79, 78, 78, 79, 79, 80, 80, 79, 84, 79, 85, 78, 86, 79,
     85, 80, 84, 79,208, 88, 79, 87, 78, 88, 79, 89, 80, 88, 84, 88, 85, 87, 86,
     88, 85, 89, 84, 88,209, 75, 78, 76, 76, 78, 75, 81, 75, 83, 76, 86, 78, 88,
     81, 89, 83, 89, 86, 88, 88, 86, 89, 83, 89, 81, 88, 78, 86, 76, 83, 75, 81,
     75,210, 81, 81, 82, 82, 83, 83, 82, 82, 81,209, 75, 80, 76, 80, 77, 81, 78,
    212, 75, 85, 76, 85, 77, 84, 78,210, 78, 80, 79, 79, 81, 79, 82, 80, 84, 82,
     85, 83, 85, 85, 84, 86, 82, 86, 81, 85, 79, 83, 78, 82, 78,210, 85, 82, 89,
    211, 85, 83, 89,207, 87, 86, 87,209, 75, 78, 76, 76, 78, 75, 81, 75, 83, 76,
     86, 78, 88, 81, 89, 83, 89, 86, 88, 88, 86, 89, 83, 89, 81, 88, 78, 86, 76,
     83, 75, 81, 75,210, 75, 82, 89,203, 82, 89, 82,215, 75, 84, 75, 81, 76, 79,
     78, 78, 81, 78, 83, 79, 86, 81, 88, 84, 89, 87, 89,215, 75, 84, 76, 82, 78,
     81, 81, 81, 83, 82, 86, 84, 88, 87, 89,203, 72, 75, 92,204, 72, 76, 92,216,
     72, 88, 92,217, 72, 89, 92,200, 72, 92, 72,200, 92, 79, 92,213, 92, 92, 92,
    203, 72, 82, 81, 74, 92,202, 72, 81, 81,202, 72, 89, 72, 90, 77, 88, 72,203,
     91, 88, 91,202, 92, 89, 92, 90, 87, 88, 92,128,210, 89, 81, 90, 82, 91, 83,
     90, 82, 89,210, 91, 81, 90, 82, 89, 83, 90, 83, 92, 82, 94, 81, 95,210, 77,
     81, 78, 82, 79, 83, 78, 82, 77,210, 89, 81, 90, 82, 91, 83, 90, 82, 89,210,
     77, 81, 78, 82, 79, 83, 78, 82, 77,210, 91, 81, 90, 82, 89, 83, 90, 83, 92,
     82, 94, 81, 95,210, 70, 81, 72, 82, 84, 83, 72, 82, 70,210, 72, 82, 78,210,
     89, 81, 90, 82, 91, 83, 90, 82, 89,205, 74, 78, 75, 77, 76, 76, 75, 76, 74,
     77, 72, 78, 71, 80, 70, 83, 70, 86, 71, 87, 72, 88, 74, 88, 76, 87, 78, 86,
     79, 82, 81, 82, 84,211, 70, 85, 71, 86, 72, 87, 74, 87, 76, 86, 78, 84, 80,
    210, 89, 81, 90, 82, 91, 83, 90, 82, 89,206, 70, 77, 77,207, 70, 77, 77,214,
     70, 85, 77,215, 70, 85, 77,209, 70, 79, 71, 78, 73, 78, 75, 79, 77, 81, 78,
     83, 78, 85, 77, 86, 75, 86, 73, 85, 71, 83, 70, 81, 70,210, 70, 82, 82,205,
     73, 87, 79,215, 73, 77, 79,219, 66, 73, 98,214, 66, 84, 68, 82, 71, 80, 75,
     79, 80, 79, 84, 80, 89, 82, 93, 84, 96, 86, 98,212, 68, 82, 72, 81, 75, 80,
     80, 80, 84, 81, 89, 82, 92, 84, 96,206, 66, 80, 68, 82, 71, 84, 75, 85, 80,
     85, 84, 84, 89, 82, 93, 80, 96, 78, 98,208, 68, 82, 72, 83, 75, 84, 80, 84,
     84, 83, 89, 82, 92, 80, 96,207, 66, 79, 98,208, 66, 80, 98,207, 66, 86, 66,
    207, 98, 86, 98,212, 66, 84, 98,213, 66, 85, 98,206, 66, 85, 66,206, 98, 85,
     98,212, 66, 82, 67, 81, 68, 80, 70, 80, 72, 81, 74, 82, 75, 83, 77, 83, 79,
     81, 81,210, 67, 81, 69, 81, 71, 82, 73, 83, 74, 84, 76, 84, 78, 83, 80, 79,
     82, 83, 84, 84, 86, 84, 88, 83, 90, 82, 91, 81, 93, 81, 95, 82, 97,209, 83,
     83, 85, 83, 87, 82, 89, 81, 90, 80, 92, 80, 94, 81, 96, 82, 97, 84, 98,208,
     66, 82, 67, 83, 68, 84, 70, 84, 72, 83, 74, 82, 75, 81, 77, 81, 79, 83, 81,
    210, 67, 83, 69, 83, 71, 82, 73, 81, 74, 80, 76, 80, 78, 81, 80, 85, 82, 81,
     84, 80, 86, 80, 88, 81, 90, 82, 91, 83, 93, 83, 95, 82, 97,211, 83, 81, 85,
     81, 87, 82, 89, 83, 90, 84, 92, 84, 94, 83, 96, 82, 97, 80, 98,210, 66, 82,
     98,201, 82, 91, 82,210, 73, 82, 91,201, 82, 91, 82,201, 79, 91, 79,201, 85,
     91, 85,218, 73, 74, 82, 90, 91,202, 73, 90, 82, 74, 91,201, 85, 73, 83, 74,
     80, 76, 79, 78, 79, 80, 80, 84, 83, 86, 84, 88, 84, 90, 83, 91, 81,201, 83,
     74, 81, 76, 80, 78, 80, 80, 81, 84, 84, 86, 85, 88, 85, 90, 84, 91, 81, 91,
     79,202, 84, 82, 79, 90, 84,202, 84, 82, 80, 90, 84,210, 72, 81, 71, 82, 70,
     83, 71, 83, 73, 82, 75, 81, 76,211, 70, 82, 71, 81, 73, 81, 75, 82, 76, 83,
     75, 82, 74,219, 70, 73, 91,206, 70, 80, 72, 80, 74, 79, 76, 77, 77, 75, 77,
     73, 75, 73, 73, 74, 71, 76, 70, 78, 70, 80, 71, 83, 72, 86, 72, 89, 71, 91,
     70,215, 84, 85, 85, 84, 87, 84, 89, 86, 91, 88, 91, 90, 90, 91, 88, 91, 86,
     89, 84, 87, 84,219, 78, 90, 79, 91, 80, 92, 79, 92, 78, 91, 77, 90, 77, 89,
     78, 88, 80, 86, 85, 84, 88, 82, 90, 80, 91, 77, 91, 74, 90, 73, 88, 73, 85,
     74, 83, 80, 79, 82, 77, 83, 75, 83, 73, 82, 71, 80, 70, 78, 71, 77, 73, 77,
     75, 78, 78, 80, 81, 85, 88, 87, 90, 90, 91, 91, 91, 92, 90, 92, 89,205, 91,
     75, 90, 74, 88, 74, 85, 75, 83, 77, 81,205, 75, 78, 77, 86, 88, 88, 90, 90,
     91,215, 78, 86, 76, 84, 75, 81, 75, 79, 76, 78, 77, 77, 80, 77, 83, 78, 85,
     80, 86, 83, 86, 85, 85, 86, 83,209, 75, 79, 77, 78, 80, 78, 83, 79, 85, 80,
     86,215, 75, 86, 83, 86, 85, 88, 86, 90, 86, 92, 84, 93, 81, 93, 79, 92, 76,
     91, 74, 89, 72, 87, 71, 84, 70, 81, 70, 78, 71, 76, 72, 74, 74, 73, 76, 72,
     79, 72, 82, 73, 85, 74, 87, 76, 89, 78, 90, 81, 91, 84, 91, 87, 90, 89, 89,
     90, 88,216, 75, 87, 83, 87, 85, 88, 86,208, 66, 80, 95,212, 66, 84, 95,216,
     73, 87, 74, 88, 75, 89, 74, 89, 73, 87, 71, 84, 70, 80, 70, 77, 71, 75, 73,
     75, 75, 76, 77, 77, 78, 79, 79, 85, 81, 87, 82, 89, 84,203, 75, 77, 77, 79,
     78, 85, 80, 87, 81, 88, 82, 89, 84, 89, 88, 87, 90, 84, 91, 80, 91, 77, 90,
     75, 88, 75, 87, 76, 86, 77, 87, 76, 88,211, 70, 76, 98,217, 70, 82, 98,204,
     81, 90, 81,203, 87, 89, 87,207, 75, 85, 89,213, 75, 79, 89,203, 79, 89, 85,
    217, 79, 75, 85,204, 91, 89, 91,
   };


static struct glfspec *fontindex = simplex_index;
static char *fonttable = simplex_table;

/*========== routines =======================================================*/

/*----------------------------------------------------------------------------*
 *  psbox.c
 *----------------------------------------------------------------------------*
 * Function: Annotate the current plot area with frame, axes,
 *       numeric labels, etc.
 *
 * XOPT (input, character): string of options for X (horizontal)
 *  axis of plot. Options are single letters, and may be in
 *  any order (see below).
 *
 * YOPT (input, character): string of options for Y (vertical)
 *  axis of plot. Coding is the same as for XOPT.
 *
 *  b : draw bottom (X) or left (Y) edge of frame
 *  c : draw top (X) or right (Y) edge of frame
 *  o : draw ticks outside viewport instead of inside.
 *  n : write Numeric labels in the conventional location
 *      below the viewport (X) or to the left of the viewport (Y)
 *  t : draw major Tick marks at the major coordinate interval
 *  s : draw minor tick marks (Subticks)
 *
 * To get a complete frame, specify BC in both XOPT and YOPT.
 *
 * Caution: the numeric label will only be drawn at a major tick mark if the
 * value can be exactly represented in 5 characters (including minus
 * sign and decimal point). If the axis runs from 1000000 to 2000000 or from
 * 0.0001 to 0.0002, say, PGBOX will be unable to label it.
 *---------------------------------------------------------------------------*/
void
psbox(char *xopt, char *yopt)
{
    extern char xflags[], yflags[];
    extern int ps_xlosc, ps_ylosc, ps_xhisc, ps_yhisc;

    /* Decode x options */
    chkoptions(xopt, xflags);

    /* Draw box X sides ?? */
    if(X_STDSIDE) gxline(ps_xlosc, ps_ylosc, ps_xhisc, ps_ylosc,'S');
    if(X_ALTSIDE) gxline(ps_xlosc, ps_yhisc, ps_xhisc, ps_yhisc,'S');
    drwxticks();

    /* Decode y options */
    chkoptions(yopt, yflags);

    /* Draw box Y sides ?? */
    if(Y_STDSIDE) gxline(ps_xlosc, ps_ylosc, ps_xlosc, ps_yhisc,'S');
    if(Y_ALTSIDE) gxline(ps_xhisc, ps_ylosc, ps_xhisc, ps_yhisc,'S');
    drwyticks();
}


/*----------------------------------------------------------------------------*
 *  pslabel (xlbl, ylbl)
 *----------------------------------------------------------------------------*
 * Function: Write labels outside the viewport.
 *
 * Arguments:
 *
 * XLBL (input, character): a label for the x-axis (centered
 *  below the viewport).
 *
 * YLBL (input, character): a label for the y-axis (centered to
 *  the left of the viewport, drawn vertically).
 *
 *---------------------------------------------------------------------------*/
void
psboxlabels(char *xlbl, char *ylbl)
{
    int length, x, y;
    extern int ps_xlosc, ps_ylosc, ps_xhisc, ps_yhisc;
    extern int ps_xchgap, ps_ychgap;

    if(xlbl != NULL){
        length = gxstrsize(xlbl, 'x')/2;
        x = ps_xlosc + (ps_xhisc - ps_xlosc)/2 - length;
        y = ps_ylosc - (int)(AXISLBLGAP * (double)ps_xchgap);
        psstrsc(x, y, xlbl);
    }

    if(ylbl != NULL){
        pssetangle(90);
        length = gxstrsize(ylbl, 'y')/2;
        x = ps_xlosc - (int)(AXISLBLGAP * (double)ps_ychgap);
        y = ps_ylosc + (ps_yhisc - ps_ylosc)/2 - length;
        psstrsc(x, y, ylbl);
        pssetangle(00);
    }
}


/*----------------------------------------------------------------------------*
 *  pstiklabel (val, clbl)
 *----------------------------------------------------------------------------*
 *
 * Function: (internal routine): Used by psbox to generate a numeric label
 * for a plot axis, using the best possible format.
 * The format is up to six characters, so the value must be in one of the
 * ranges -99999->999999 (integer), -999.9->9999.9 (multiple of 0.1),
 * -99.99->999.99 (multiple of 0.01), or 0.0001->0.9999 (multiple of 0.001).
 *
 * Arguments:
 *
 * float val : the numeric value to be formatted.
 *
 * char *clbl : pointer to a character array to receive the
 *  formatted label. The label is left-justified in the
 *  string.
 *
 *---------------------------------------------------------------------------*/
static int
pstiklabel(double val, char *clbl)
{
    double absval, eps;
    int nchar;
    char *s;

    absval = fabs(val);

    if(absval>1000000.0) {
        strcpy(clbl, " ");
        return(1);
    }
    if(absval<1e-7) {
        strcpy(clbl, "0.0");
        return(1);
    }

    if(val<100.0 && val>-10.0){
        sprintf(clbl, "%.3f", val);
        goto trimit;
    }
    else if(val<1000.0 && val>-100.0){
        sprintf(clbl, "%.2f", val);
        goto trimit;
    }
    else if(val<10000.0 && val>-1000.0){
        sprintf(clbl, "%.1f", val);
        goto trimit;
    }
    else if(val<1000000.0 && val>-100000.0){
        sprintf(clbl, "%.0f", val);
        return(strlen(clbl));
    }
    else    {
        strcpy(clbl, " ");
        return(1);
    }

    trimit: /* trim off trailing zeros */
    nchar = strlen(clbl);
    s = clbl+nchar-1;
    while(*(s)=='0' && *(s-1)!='.') {
        *s-- = '\0';
        nchar--;
    }
    return(nchar);
 }

static void
drwxticks(void)
{
    int i, j, k;
    int nsub_int, nsub_inc;
    int tik_long, tik_shrt, tik_size;
    int place, part, tmp1, tmp2;
    int bigstepsc;
    int t_1st;
    double bigstepwc;
    double value;

    extern char xflags[], clbl[];
    extern int ps_xhisc, ps_xlosc, ps_yhisc, ps_ylosc;
    extern int  ps_ytxtscal, ps_xchgap;
    extern double ps_xscale, ps_xoffset;

    /* Choose x tick lengths */
    tik_long = GLFTALL * ps_ytxtscal;       /* tik is in y direction */
    tik_long = (tik_long*7)/10;
    tik_shrt = tik_long/2;

    /* Choose x major tick intervals */
    bigstepsc =  GLFWIDE * ps_xtxtscal * 7;
    bigstepwc = (double)bigstepsc / ps_xscale;  /* convert to user coord */
    bigstepwc = (double)psround(bigstepwc, &nsub_int);
                            /* make it a 'nice' number */
    bigstepsc = (int)(bigstepwc * ps_xscale);
    if(!X_SMLTIC) nsub_int = 1;
    nsub_inc = 10/nsub_int;

    /* Draw x ticks ?? */
    if(X_BIGTIC || X_SMLTIC){
        t_1st = (int)((ps_xlowc-(bigstepwc/2))/bigstepwc);
        k = 10;

        for(i=0;; i++)  {       /* counting major ticks */
            value = (double)(t_1st + i) * bigstepwc;
            place = (int)(ps_xoffset + (value * ps_xscale));

            for(j=0, tik_size=tik_long; j<k; j+=nsub_inc, tik_size=tik_shrt){
                part = (int)((double)bigstepsc*lintbl[j]);

                /* are we in the plot area ?? */
                if((place+part)<ps_xlosc) continue;     /* not yet */
                if((place+part)>ps_xhisc){              /* finished */
                    pssetltype(0);
                    return;
                }

                /* draw minor ticks as well as major ? */
                if((j==00 || X_SMLTIC) && ((place+part)>ps_xlosc
                    &&  (place+part)<ps_xhisc)) {
                    /* do a mid-size tick ?? */
                    if(j==5) tik_size += tik_size/2;

                    /* do bottom */
                    if(X_STDSIDE) gxline(place+part,  ps_ylosc,
                        place+part, (ps_ylosc + tik_size),'S');

                    /* do right */
                    if(X_ALTSIDE) gxline(place+part,
                        ps_yhisc, place+part, (ps_yhisc - tik_size),'S');
                }

                /* Write x labels */
                if(j==0 && X_STDLBL){
                    pstiklabel((double)value, clbl);

                    tmp1 = place - gxstrsize(clbl, 'x')/2;
                    if(X_STDLBL) tmp2 = ps_ylosc - ps_xchgap;
                    psstrsc(tmp1, tmp2, clbl);
                }
            }
        }
    }
}

static void
drwyticks(void)
{
    int i, j, k;
    int nsub_int, nsub_inc;
    int tik_long, tik_shrt, tik_size;
    int place, part, tmp1, tmp2;
    int bigstepsc;
    int t_1st;
    double bigstepwc;
    double value;

    extern char yflags[], clbl[];
    extern int ps_xhisc, ps_xlosc, ps_yhisc, ps_ylosc;
    extern int  ps_ytxtscal, ps_ychgap;
    extern double ps_yscale, ps_yoffset;

    /* Choose Y tick lengths */
    tik_long = GLFTALL * ps_xtxtscal;       /* tik is in x direction */
    tik_long = (tik_long*7)/10;
    tik_shrt = tik_long/2;

    /* Choose Y major tick intervals */

    bigstepsc = GLFWIDE * ps_ytxtscal * 7;
    bigstepwc = (double)bigstepsc / ps_yscale;  /* convert to user coord */
    bigstepwc = (double)psround(bigstepwc, &nsub_int);
    /* make it a 'nice' number */
    bigstepsc = (int)(bigstepwc * ps_yscale);
    if(!Y_SMLTIC) nsub_int = 1;
    nsub_inc = 10/nsub_int;

    /* Draw y ticks ?? */
    if(Y_BIGTIC || Y_SMLTIC){
        t_1st = (int)((ps_ylowc-(bigstepwc/2)) /bigstepwc);
        k = 10;

        pssetangle(90);
        for(i=0;; i++)  {               /* counting major ticks */
            value = (double)(t_1st + i) * bigstepwc;
            place = (int)(ps_yoffset + (value * ps_yscale));

            for(j=0, tik_size=tik_long; j<k; j+=nsub_inc, tik_size=tik_shrt){
                part = (int)((double)bigstepsc*lintbl[j]);

                /* are we in the plot area ?? */
                if((place+part)<ps_ylosc)       /* not yet */
                    continue;
                if((place+part)>ps_yhisc){      /* gone past */
                    pssetltype(0);
                    pssetangle(00);
                    return;
                }

                /* draw minor ticks as well as major ? */
                if((j==00 || Y_SMLTIC) && ((place+part)>ps_ylosc
                    &&  (place+part)<ps_yhisc)) {
                    /* do a mid-size tick ?? */
                    if(j==5) tik_size += tik_size/2;

                    /* do left side ? */
                    if(Y_STDSIDE) gxline(ps_xlosc, place+part,
                        (ps_xlosc + tik_size), place+part,'S');

                    /* do right */
                    if(Y_ALTSIDE) gxline(ps_xhisc, place+part,
                        (ps_xhisc - tik_size), place+part,'S');
                }

                /* Write y labels */
                if(j==0 && Y_STDLBL){
                    pstiklabel((double)value, clbl);

                    if(Y_STDLBL) tmp1 = ps_xlosc - ps_ychgap;
                    tmp2 = place - gxstrsize(clbl, 'y')/2;
                    psstrsc(tmp1, tmp2, clbl);
                }
            }
        }
    }
}

static void
chkoptions(char *str, char *opflgs)
{
    int i;
    char *sp;
    extern char optsymbols[];

    for(i=0; i< NOPTIONS; i++){
        sp = str;
        opflgs[i] = 00;
        while(*sp != '\00') {
            if(*sp++ == optsymbols[i]){
                opflgs[i] = 1;
                break;
            }
        }
    }
}


/* connect the points specified by x & y. */
void
psconnect(double *x, double *y, int npts)
{
    int i;
    int xnew, ynew;
    int xold, yold;

    extern double ps_xoffset, ps_yoffset, ps_xscale, ps_yscale;

    xold = (int)(ps_xoffset + (*x++ * ps_xscale));
    yold = (int)(ps_yoffset + (*y++ * ps_yscale));
    npts--;

    for (i=0; i<npts; i++) {
        xnew = (int)(ps_xoffset + (*x++ * ps_xscale));
        ynew = (int)(ps_yoffset + (*y++ * ps_yscale));
        pslinesc(xold, yold, xnew, ynew);
        xold = xnew;
        yold = ynew;
    }
}


/*  connect the points specified by x & y in hisogram style. 
 *  equal x steps are assumed */
void 
pshisto(double *x, double *y, int npts)
{
    int i;
    int xnew, ynew;
    int xold, yold;
    int hafbin;
    int x1, x2;

    extern double ps_xoffset, ps_yoffset, ps_xscale, ps_yscale;

    x1 = (int)(ps_xoffset + (*x * ps_xscale));
    x2 = (int)(ps_xoffset + (*(x+1) * ps_xscale));
    hafbin = (x2-x1) / 2;

    xold = (int)(ps_xoffset + (*x++ * ps_xscale));
    yold = (int)(ps_yoffset + (*y++ * ps_yscale));
    xnew = xold + hafbin;
    npts--;
    pslinesc(xold, yold, xnew, yold);       /* draw first half segment */
    xold = xnew;

    for (i=0; i<npts; i++) {
        xnew = hafbin + (int)(ps_xoffset + (*x++ * ps_xscale));
        ynew = (int)(ps_yoffset + (*y++ * ps_yscale));
        pslinesc(xold, yold, xold, ynew);   /* draw vertical segment */
        pslinesc(xold, ynew, xnew, ynew);   /* horiz segment centered at x */
        xold = xnew;
        yold = ynew;
    }
}


/*========== no_device routines =============================================*/
static int no_open(void)    { return(0); }
static int no_plotmode(void) { return(1000); }  /* returns aspect-ratio * 1000 */
static int no_line(void)    {; }
static int no_dot(void)     { return(-1); }
static int no_ltype(void)   { return(-1); }
static int no_lwidth(void) { return(-1); }
static int no_erase(void)  {; }
static int no_cursor(void) { return(-1); }
static int no_color(void)  {; }
static int no_gflush(void) {; }
static int no_textmode(void) {; }
static int no_close(void)  {; }

/*========== ps_device routines =============================================*/
void
psopen(char *devname, char mode, char *fname)
{
    char dev = *devname;

    if(dev == 'x' || dev == 'X')
        ps_device = DEV_X11;

    DEV_OPEN(mode, fname);
    ps_isopen = 1;
}


/* change device from textmode to plotmode .
 *  The plot device is expected to return an integer equal
 *  to its the current aspect ratio * 1000 */
void
psplotmode(void)
{
    int asp;

    if(ps_isopen == 0) psopen("x11", 0, NULL);

    asp = DEV_PLOTMODE();
    pstxtaspect((double)asp / 1000);
}


/*  draw a line in screen coordinates from (x0,y0) to (x1,y1)
 *  clipping the line to the viewport if ps_clip is TRUE
 *  returns FALSE if the line was outside the plot area or the
 *  screen limits.
 */
int
pslinesc(int x0, int y0, int x1, int y1)
{
    int outcode0, outcode1, r;

    /* if no plot area clipping, only check screen limits
     * otherwise check plot area limits  */

    if(!ps_clip)
        return(gxline(x0, y0, x1, y1, 'S'));
    else
        return(gxline(x0, y0, x1, y1, 'A'));
}


/* Set a linetype. Linetypes supported in hardware are set to be -ve  */
int
pssetltype(int arg)
{
    int oldltype;
    extern int ps_ltype;

    if((oldltype = DEV_LTYPE(arg)) >= 0){
        oldltype = ps_ltype;
        ps_ltype = -arg;
    }
    else{
        oldltype = ps_ltype;
        ps_ltype = arg & 07;
    }

    return(oldltype);
}


/* Set the linewidth. Linewidth is supported in hardware only  */
/*
int
pssetlwidth(int arg)
{
    static int oldlwidth = 0;
    int r;

    if((r = DEV_LWIDTH(arg)) >= 0){
        r = oldlwidth;
        oldlwidth = arg;
    }
    return(r);
}
*/


void
pserase(void)
{
    DEV_ERASE();
}


/*----------------------------------------------------------------------------*
 *  pscurse (x, y, ch)
 *----------------------------------------------------------------------------*
 *  Read the cursor position and a character typed by the user.
 * The position is returned in world coordinates.
 * PSCURSE positions the cursor at the position specified,
 * allows the user to move the cursor using the joystick or
 * arrow keys or whatever is available on the device. When he has
 * positioned the cursor, the user types a single character on his
 * keyboard; PSCURSE then returns this character and the new cursor
 * position (in world coordinates).
 *
 * Arguments:
 *
 * *x : the world x-coordinate of the cursor. (input/output)
 * *y : the world y-coordinate of the cursor. (input/output)
 * *ch: the character typed by the user. (output)
 *---------------------------------------------------------------------------*/

/*
void
psrdcurswc(double *x, double *y, int *ch)
{
    int xsc, ysc, rc;

    extern double ps_xoffset, ps_yoffset, ps_xscale, ps_yscale;

    xsc = (int)(ps_xoffset + (*x * ps_xscale));
    ysc = (int)(ps_yoffset + (*y * ps_yscale));

    rc = DEV_RDCURSOR(&xsc, &ysc);

    *x = (xsc - ps_xoffset)/ps_xscale;
    *y = (ysc - ps_yoffset)/ps_yscale;

    if(rc <= 0){
        puts("CALL TO pscurse IGNORED FOR DEVICE");
        *ch = '?';
    }
    else{
        *ch = (char)rc;
    }
}
*/


int
pssetcolor(int i)
{
    extern int ps_color;
    int oldcolor = ps_color;

    ps_color = i&07;
    DEV_COLOR(i);

    return(oldcolor);
}

void
psgflush(void)
{
    DEV_GFLUSH();
}

void
pstextmode(void)
{
    DEV_TEXTMODE();
}

void
psclose(void)
{
    DEV_CLOSE();
}


/*========== helper routines ================================================*/


/*----------------------------------------------------------------------------*
 *  psround (v, nsub)
 *----------------------------------------------------------------------------*
 * Usage: psround(v, nsub);
 *
 * Function: (internal routine): used by psbox, returns the smallest 
 * "nice" number larger than v, a "nice" number being 1, 2 or 5 times 
 * a power of 10. If v is negative, psround(v) = -psround(abs(v)).
 * eg. psround(8.7) = 10.0, psround(-0.4) = -0.5
 *
 * Arguments:
 *
 *  float v; the number to be rounded. Note: it may not be zero!
 *
 *  float *nsub; a suitable number of subdivisions for
 *  subdividing the "nice" number into (2 or 5) is returned here .
 *
 * Returns: the "nice" number;
 *
 *---------------------------------------------------------------------------*/


static double
psround(double v, int *nsub)
{
    int  i, ilog;
    double frac, pwr, vlog, vabs, nice_val;
    extern double nice[];

    vabs   = fabs(v);
    vlog = log10(vabs);
    ilog = (int)vlog;
    if(vlog<0.0) ilog--;
    pwr  = pow(10.0, (double)ilog);
    frac = vabs/pwr;

    for(i=0; i<=2; i++){
        if(frac <= nice[i]){
            *nsub = (int)nice[i];
            if(v > 0.0)
                nice_val =  pwr * nice[i];
            else
                nice_val = -pwr * nice[i];
            break;
        }
    }
    return(nice_val);
}


/* set x & y scale factors, & magnified font width & height */
static void
settextsizes(void) 
{
    extern double ps_textaspect, ps_textscale;
    extern int ps_xtxtscal, ps_ytxtscal;

    if(ps_textaspect > 1.0) {
        ps_xtxtscal = (int)((double)GLFMAG * ps_textscale);
        ps_ytxtscal = (int)(((double)GLFMAG * ps_textscale) / ps_textaspect);
    } 
    else if (ps_textaspect < 1.0) {
        ps_xtxtscal = (int)((double)GLFMAG * ps_textscale * ps_textaspect);
        ps_ytxtscal = (int)((double)GLFMAG * ps_textscale);
    } 
    else{
        ps_xtxtscal = (int)((double)GLFMAG * ps_textscale);
        ps_ytxtscal = ps_xtxtscal;
    }

    ps_xchgap  = ps_ytxtscal * GLFTALL;
    ps_ychgap  = ps_xtxtscal * GLFTALL;
}

/* Set character size */
static void
pstxtscale(double arg)
{
    extern double ps_textscale;

    ps_textscale = (double)arg;
    settextsizes();
}

/* Set character aspect ratio */
static void
pstxtaspect(double arg)
{
    extern double ps_textaspect;

    ps_textaspect = (double)arg;
    settextsizes();
}


/* set the angle for text, increments of 90 degrees only */
static void
pssetangle(int arg)
{
    extern int ps_angle, ps_sina, ps_cosa;

    ps_angle = arg/90;

    switch(ps_angle){
    case 0 :
        ps_sina = 0;
        ps_cosa = 1;
        break;
    case 1 :
        ps_sina = 1;
        ps_cosa = 0;
        break;
    case 2 :
        ps_sina =  0;
        ps_cosa = -1;
        break;
    case 3 :
        ps_sina = -1;
        ps_cosa =  0;
        break;
    }
}


/* define the current margins between the window edges and the plot area,
 * in terms of character height. Defaults are 5 .
 */
int 
pssetmarg(double bot, double top, double left, double right)
{
    double t1, t2, xchtall, ychtall;
    extern double ps_textaspect;
    extern int ps_leftmarg, ps_rightmarg, ps_botmarg, ps_topmarg;

    if(bot < 0.0 || top < 0.0 || left < 0.0 || right < 0.0
        || (t1=bot+top) >= 40.0 || (t2=left+right) >= 40.0){
        fprintf(stderr, "PSSETMARG : illegal margins definition\n");
        return(FALSE);
    }

    if(ps_textaspect > 1.0){                    /* Y > X */
        xchtall = CHTALL;
        ychtall = CHTALL / ps_textaspect;
    } 
    else if (ps_textaspect < 1.0){              /* Y < X */
        xchtall = CHTALL * ps_textaspect;
        ychtall = CHTALL;
    } 
    else {                                  /* ps_textaspect = 1.0 */ 
        xchtall = ychtall = (double)CHTALL;
    }

    ps_leftmarg = (int)(left * xchtall);    
    ps_rightmarg = (int)(right * xchtall);
    ps_botmarg = (int)(bot * ychtall);
    ps_topmarg = (int)(top * ychtall);

    return(TRUE);
}


/* adjust scale & offset factors, used in pssetarea() & pssetwcs() */
static void
pssetscale(void)
{
    extern int ps_xlosc, ps_ylosc, ps_xhisc, ps_yhisc;
    extern double ps_xlowc, ps_ylowc, ps_xhiwc, ps_yhiwc;
    extern double ps_xoffset, ps_yoffset, ps_xscale, ps_yscale;

    ps_xscale  = (double)(ps_xhisc - ps_xlosc)/(ps_xhiwc - ps_xlowc);
    ps_yscale  = (double)(ps_yhisc - ps_ylosc)/(ps_yhiwc - ps_ylowc);

    ps_xoffset = (double)ps_xlosc - (ps_xlowc * ps_xscale); 
    ps_yoffset = (double)ps_ylosc - (ps_ylowc * ps_yscale); 
}


/* define the current plot window in normalized coordinates, 0.0 -> 1.0 
 * this should really be called pssetwindow. 
 * The size sequence is : screen (full device area), window (sub-area),
 * plot area (window - margins).
 */
int 
pssetarea(double xlo, double ylo, double xhi, double yhi)
{
    int xlosc, ylosc, xhisc, yhisc;
    extern int ps_xlosc, ps_ylosc, ps_xhisc, ps_yhisc;
    extern int ps_leftmarg, ps_rightmarg, ps_botmarg, ps_topmarg;

    xlosc = (int)(xlo*SC_MAX);  
    xhisc = (int)(xhi*SC_MAX);
    ylosc = (int)(ylo*SC_MAX);
    yhisc = (int)(yhi*SC_MAX);

    if(xlosc>=xhisc || ylosc>=yhisc || xlosc < 00 || ylosc < 00
        || xhisc>SC_MAX || yhisc>SC_MAX){
        fprintf(stderr, "PSSETAREA : illegal area definition\n");
        return(FALSE);
    }

    /* subtract margins to find the plot area */
    xlosc += ps_leftmarg;   
    xhisc -= ps_rightmarg;
    ylosc += ps_botmarg;
    yhisc -= ps_topmarg;

    /* set screen coords of plot area, set scales & offsets */
    ps_xlosc = xlosc;   
    ps_xhisc = xhisc;
    ps_ylosc = ylosc;
    ps_yhisc = yhisc;
    pssetscale();

    return(TRUE);
}


/* define world coordinate system for the current plot area. see ps_setarea() */
void 
pssetwcs(double xlo, double ylo, double xhi, double yhi)
{
    extern double ps_xlowc, ps_ylowc, ps_xhiwc, ps_yhiwc;

    /* set user coords of graphics area */
    ps_xlowc = (double)xlo; 
    ps_xhiwc = (double)xhi;
    ps_ylowc = (double)ylo;
    ps_yhiwc = (double)yhi;
    pssetscale();
}


/* reset variables to default conditions */
void
psreset(void)
{
    if(ps_isopen == 00) psopen("x11", 0, NULL);

    pssetmarg(2.0, 2.0, 2.0, 2.0);
    pssetarea(0.0, 0.0, 1.0, 1.0);
    pssetwcs(0.0, 0.0, 1.0, 1.0);
    pssetangle(00);
    pstxtscale(1.0);
    pssetltype(00);
    pssetcolor(1);
}


/* write str at xpos, ypos in screen coordinates */
static void
psstrsc(int xpos, int ypos, char *str)
{
    extern int ps_clip;
    int dum1, dum2;
    int oldclip = ps_clip;

    ps_clip = FALSE;

    gxstring(xpos, ypos, str, TRUE, &dum1, &dum2);
    ps_clip = oldclip;
}


/*===========================================================================*/

/* check dx & dy of str in screen coordinates with current text-size & angle */
static int
gxstrsize(char *str, char axis)
{
    int str_dx, str_dy;

    gxstring(00L, 00L, str, FALSE, &str_dx, &str_dy);

    if(axis == 'x') 
        return(str_dx);
    else    
        return(str_dy);
}


/*  lo-level string routine
 *  maybe draw str at xpos, ypos in screen coordinates
 *  return sizes, dx & dy
 *
 *  option escape sequences:
 *  \{            superscript
 *  \[            subscript
 *  \n            back to normal
 *  \}            end superscript
 *  \]            end subscript
 *  \b            move back to last pos'n
 *  \\            maybe a drawable '\'
 *
 */
static void
gxstring(int xpos, int ypos, char *str, int drwflg, 
    int *str_dx, int *str_dy)
{
    int i;
    int nstrokes  = 0;
    int newstroke = 0;
    int ind;
    int tbl_x, tbl_y; 
    int radj, ladj; 
    int cur_glf;
    int glfoffset = 0;
    int x_drw1, y_drw1, x_drw2, y_drw2; 
    int cur_xpos, cur_ypos;
    int old_xpos, old_ypos;
    int  charw;
    int gxscl, gyscl;    /* glyph x & y, scale depends on angle for aspect */
    int wxscl, wyscl;   /* width x & y = screen x & y */

    extern int ps_xtxtscal, ps_ytxtscal, ps_cosa, ps_sina;
    extern struct glfspec *fontindex;
    extern char *fonttable;

    gxscl = ps_xtxtscal * ps_cosa +  ps_ytxtscal * ps_sina;
    gyscl = ps_xtxtscal * ps_sina +  ps_ytxtscal * ps_cosa;
    wxscl = ps_xtxtscal;
    wyscl = ps_ytxtscal;

    old_xpos = cur_xpos = xpos;
    old_ypos = cur_ypos = ypos;

    while(*str != '\00') {
        if (*str == '\\') {
            escapechar:
            switch (*++str) {
                case '{':           /* superscript */
                gxscl = (int)((double)gxscl*0.6);
                gyscl = (int)((double)gyscl*0.6);
                wxscl = (int)((double)wxscl*0.6);
                wyscl = (int)((double)wyscl*0.6);
                glfoffset += SUBOFF;
                break;
            case '[':           /* supscript */
                gxscl = (int)((double)gxscl*0.6);
                gyscl = (int)((double)gyscl*0.6);
                wxscl = (int)((double)wxscl*0.6);
                wyscl = (int)((double)wyscl*0.6);
                glfoffset -= SUBOFF;
                break;
            case 'n':           /* back to normal */
            case '}':           /* end superscript */
            case ']':           /* end subscript */
                gxscl = ps_xtxtscal * ps_cosa +  ps_ytxtscal * ps_sina;
                gyscl = ps_xtxtscal * ps_sina +  ps_ytxtscal * ps_cosa;
                wxscl = ps_xtxtscal;
                wyscl = ps_ytxtscal;
                glfoffset = 0;
                break;
            case 'b':           /* move back to last pos'n */
                cur_xpos = old_xpos;
                cur_ypos = old_ypos;
            }
            /* allow for consecutive escapes */
            if(*str != '\\' && *(str+1) == '\\') { 
                str++; 
                goto escapechar; 
            }
            if(*str != '\\') str++;     /* maybe a drawable '\' */
        }

        ind = *str++ - 1;
        cur_glf  = (fontindex+ind)->offset; /* start of glyph */
        nstrokes = (fontindex+ind)->npoints;    /* no. of coordinate PAIRS */
    
        radj = (int)((fontindex+ind)->radj);    /* radj */
        ladj = (int)((fontindex+ind)->ladj);    /* ladj */
        charw = ladj + radj;

        if(drwflg){
            for(i=0; i<nstrokes; i++) {
                tbl_x = *(fonttable+cur_glf++);
                tbl_y = *(fonttable+cur_glf++);

                if (tbl_x & 128){           /* new stroke - just move */
                    newstroke = TRUE;
                    tbl_x &= 127;
                }

                tbl_x =  gxscl * (tbl_x - 'R' + radj);
                tbl_y = -gyscl * (tbl_y - 'R' - glfoffset);
                x_drw2 = cur_xpos + (int)(tbl_x*ps_cosa - tbl_y*ps_sina);
                y_drw2 = cur_ypos + (int)(tbl_x*ps_sina + tbl_y*ps_cosa);

                if(newstroke){
                    x_drw1 = x_drw2;
                    y_drw1 = y_drw2;
                    newstroke = FALSE;
                }
                else{
                    gxline(x_drw1, y_drw1, x_drw2, y_drw2,'S');
                    x_drw1 = x_drw2;
                    y_drw1 = y_drw2;
                }
            }
        }                                   /* end of if(drwflg) */

        old_xpos = cur_xpos;
        old_ypos = cur_ypos;
        cur_xpos += (int)(charw * ps_cosa * wxscl);
        cur_ypos += (int)(charw * ps_sina * wyscl);
    }
    *str_dx = cur_xpos - xpos; 
    *str_dy = cur_ypos - ypos;
}

/* draw a line from (x0,y0) to (x1,y1) clipping the line to screen limits */
/* option 'S' or to the plot-area limits option 'A' */


/* check if point xend, yend is within current plot area (limitflag = 'A') *
 * check if point xend, yend is within screen limits (limitflag = 'S')     *
 * return flag indicating limits crossed                                   */

static int 
gxline(int x0, int y0, int x1, int y1, char limitflag)
{
    int outcode0, outcode1, r;
    int dx, dy;
    int xlo, xhi, ylo, yhi;

    extern int ps_xlosc, ps_ylosc, ps_xhisc,ps_yhisc;

    if(limitflag == 'A'){ 
        xlo = ps_xlosc; xhi = ps_xhisc; 
        ylo = ps_ylosc; yhi = ps_yhisc;
    }
    else{ 
        xlo = 00; xhi = SC_MAX; 
        ylo = 00; yhi = SC_MAX;
    }

    dx = x1 - x0;
    dy = y1 - y0;

    while(1){
        outcode0 = 0;  
        outcode1 = 0;
        if(x0 < xlo) outcode0 |= OUT_LEFT;          /* left */
        if(x0 > xhi) outcode0 |= OUT_RIGHT;         /* right */
        if(y0 < ylo) outcode0 |= OUT_BOT;           /* bottom */
        if(y0 > yhi) outcode0 |= OUT_TOP;           /* top */

        if(x1 < xlo) outcode1 |= OUT_LEFT;          /* left */
        if(x1 > xhi) outcode1 |= OUT_RIGHT;         /* right */
        if(y1 < ylo) outcode1 |= OUT_BOT;           /* bottom */
        if(y1 > yhi) outcode1 |= OUT_TOP;           /* top */

        if((outcode0 & outcode1))                   /* both points out */
            return(FALSE);
        if(!(outcode0 | outcode1))                  /* both points in */
            break;

        /* move P0 to the intersect point                               */
        /* y = y1 + slope * (x - x1)    x = x1 + (1/slope) * (y - y1)   */

        if(outcode0 & OUT_TOP) {            /* cut line at top of window */
            x0 += dx * (yhi - y0) / dy;
            y0 = yhi;
        }
        else if(outcode0 & OUT_BOT) {       /* cut line at bottom of window */
            x0 += dx * (ylo - y0) / dy;
            y0 = ylo;
        }
        else if(outcode0 & OUT_RIGHT) {     /* cut line at right of window */
            y0 += dy * (xhi - x0) / dx;
            x0 = xhi;
        }
        else if(outcode0 & OUT_LEFT) {      /* cut line at left of window */
            y0 += dy * (xlo - x0) / dx;
            x0 = xlo;
        }

        if(outcode1 & OUT_TOP) {            /* cut line at top of window */
            x1 += dx * (yhi - y1) / dy;
            y1 = yhi;
        }
        else if(outcode1 & OUT_BOT) {       /* cut line at bottom of window */
            x1 += dx * (ylo - y1) / dy;
            y1 = ylo;
        }
        else if(outcode1 & OUT_RIGHT) {     /* cut line at right of window*/
            y1 += dy * (xhi - x1) / dx;
            x1 = xhi;
        }
        else if(outcode1 & OUT_LEFT) {      /* cut line at left of window */
            y1 += dy * (xlo - x1) / dx;
            x1 = xlo;
        }
    }

    DEV_LINE((int)x0, (int)y0, (int)x1, (int)y1);

    return(TRUE);
}
